from __future__ import annotations

from typing import Any

import numpy as np
from pydantic import BaseModel, Field
from pymatgen.analysis.elasticity.elastic import ElasticTensor, ElasticTensorExpansion
from pymatgen.analysis.elasticity.strain import Deformation, Strain
from pymatgen.analysis.elasticity.stress import Stress
from pymatgen.core.structure import Structure
from pymatgen.core.tensors import TensorMapping
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer

from emmet.core.material_property import PropertyDoc
from emmet.core.math import Matrix3D, MatrixVoigt
from emmet.core.settings import EmmetSettings
from emmet.core.types.enums import TaskState
from emmet.core.types.pymatgen_types.structure_adapter import StructureType
from emmet.core.types.typing import IdentifierType

SETTINGS = EmmetSettings()


class ElasticTensorDoc(BaseModel):
    raw: MatrixVoigt | None = Field(
        None,
        description="Elastic tensor corresponding to structure orientation (GPa)",
    )
    ieee_format: MatrixVoigt | None = Field(
        None,
        description="Elastic tensor corresponding to IEEE orientation (GPa)",
    )


class ComplianceTensorDoc(BaseModel):
    raw: MatrixVoigt | None = Field(
        None,
        description="Compliance tensor corresponding to structure orientation (TPa^-1)",
    )
    ieee_format: MatrixVoigt | None = Field(
        None,
        description="Compliance tensor corresponding to IEEE orientation (TPa^-1)",
    )


class BulkModulus(BaseModel):
    voigt: float | None = Field(None, description="Bulk modulus Voigt average (GPa)")
    reuss: float | None = Field(None, description="Bulk modulus Reuss average (GPa)")
    vrh: float | None = Field(
        None, description="Bulk modulus Voigt-Reuss-Hill average (GPa)"
    )


class ShearModulus(BaseModel):
    voigt: float | None = Field(None, description="Shear modulus Voigt average (GPa)")
    reuss: float | None = Field(None, description="Shear modulus Reuss average (GPa)")
    vrh: float | None = Field(
        None, description="Shear modulus Voigt-Reuss-Hill average (GPa)"
    )


class SoundVelocity(BaseModel):
    transverse: float | None = Field(
        None, description="Transverse sound velocity (SI units)"
    )
    longitudinal: float | None = Field(
        None, description="Longitudinal sound velocity (SI units)"
    )
    snyder_acoustic: float | None = Field(
        None, description="Snyder's acoustic sound velocity (SI units)"
    )
    snyder_optical: float | None = Field(
        None, description="Snyder's optical sound velocity (SI units)"
    )
    snyder_total: float | None = Field(
        None, description="Snyder's total sound velocity (SI units)"
    )


class ThermalConductivity(BaseModel):
    clarke: float | None = Field(
        None, description="Clarke's thermal conductivity (SI units)"
    )
    cahill: float | None = Field(
        None, description="Cahill's thermal conductivity (SI units)"
    )


class FittingData(BaseModel):
    """
    Data used to fit the elastic tensor.

    Note, this only consists of the explicitly calculated primary data.
    With the data here, one can redo the fitting to regenerate the elastic data,
    e.g. using `ElasticityDoc.from_deformations_and_stresses()`.
    """

    # data of strained structures
    deformations: list[Matrix3D] = Field(
        description="Deformations corresponding to the strained structures"
    )
    strains: list[Matrix3D] = Field(
        description="Lagrangian strain tensors applied to structures"
    )
    cauchy_stresses: list[Matrix3D] = Field(
        description="Cauchy stress tensors on strained structures"
    )
    second_pk_stresses: list[Matrix3D] = Field(
        description="Second Piola-Kirchhoff stress tensors on structures"
    )
    deformation_tasks: list[IdentifierType] | None = Field(
        None,
        description="Deformation task ids corresponding to the strained structures",
    )
    deformation_dir_names: list[str] | None = Field(
        None, description="Paths to the running directories of deformation tasks"
    )

    # data of equilibrium structure
    equilibrium_cauchy_stress: Matrix3D | None = Field(
        None, description="Cauchy stress tensor of the relaxed structure"
    )
    optimization_task: IdentifierType | None = Field(
        None, description="Optimization task corresponding to the relaxed structure"
    )
    optimization_dir_name: str | None = Field(
        None, description="Path to the running directory of the optimization task"
    )

    # derived strains stresses
    num_total_strain_stress_states: int | None = Field(
        None,
        description="Number of total strain--stress states used for fitting, i.e. the "
        "sum of explicitly calculated deformations and derived deformations from "
        "symmetry.",
    )


class CriticalMessage(BaseModel):
    FITTING: str = "Critical: cannot fit elastic tensor. Error: {}."
    COMPLIANCE: str = (
        "Critical: cannot invert elastic tensor to get compliance tensor. Error: {}."
    )
    STRAIN_RANK: str = (
        "Critical: insufficient number of valid strains. Expect the matrix of all "
        "strains to be of rank 6; got rank {}."
    )
    N_STATES: str = (
        "Critical: Expect 24 total (primary plus derived) strain--stress states "
        "to fit the data; got {}."
    )


class WarningMessage(BaseModel):
    NEGATIVE_EIGVAL: str = (
        "Elastic tensor has negative eigenvalue(s), indicating that the structure is "
        "mechanically unstable."
    )  # https://doi.org/10.1103/PhysRevB.90.224104
    NEGATIVE_MODULUS: str = (
        "Negative modulus: {} {} is {}, indicating that the structure is mechanically "
        "unstable."
    )
    SMALL_MODULUS: str = "Small modulus. {} {} is {}; smaller than {}."
    LARGE_MODULUS: str = "Large modulus. {} {} is {}; larger than {}."
    LARGE_YOUNG_MODULUS: str = "Large Young's modulus {}; larger than {}."
    LARGE_COMPONENT: str = "Elastic tensor has component larger than {}."


class ElasticityDoc(PropertyDoc):
    property_name: str = "elasticity"

    structure: StructureType | None = Field(
        None,
        description="Structure to compute the elasticity",
    )

    order: int = Field(
        default=2, description="Order of the expansion of the elastic tensor"
    )

    elastic_tensor: ElasticTensorDoc | None = Field(None, description="Elastic tensor")

    compliance_tensor: ComplianceTensorDoc | None = Field(
        None, description="Compliance tensor"
    )

    # derived properties
    bulk_modulus: BulkModulus | None = Field(None, description="Bulk modulus")
    shear_modulus: ShearModulus | None = Field(None, description="Shear modulus")
    sound_velocity: SoundVelocity | None = Field(None, description="Sound velocity")
    thermal_conductivity: ThermalConductivity | None = Field(
        None, description="Thermal conductivity"
    )
    young_modulus: float | None = Field(
        None, description="Young's modulus (SI units)", alias="youngs_modulus"
    )
    universal_anisotropy: float | None = Field(
        None, description="Universal elastic anisotropy"
    )
    homogeneous_poisson: float | None = Field(
        None, description="Homogeneous Poisson ratio"
    )
    debye_temperature: float | None = Field(
        None, description="Debye temperature (SI units)"
    )

    fitting_data: FittingData | None = Field(
        None, description="Data used to fit the elastic tensor"
    )

    fitting_method: str | None = Field(
        None, description="Method used to fit the elastic tensor"
    )

    state: TaskState | None = Field(
        None,
        description="State of the fitting/analysis: `successful` or `failed`",
    )

    @classmethod
    def from_deformations_and_stresses(
        cls,
        structure: Structure,
        deformations: list[Deformation],
        stresses: list[Stress],
        material_id: IdentifierType | None = None,
        deformation_task_ids: list[IdentifierType] | None = None,
        deformation_dir_names: list[str] | None = None,
        equilibrium_stress: Stress | None = None,
        optimization_task_id: IdentifierType | None = None,
        optimization_dir_name: str | None = None,
        fitting_method: str = "finite_difference",
        **kwargs,
    ):
        """
        Fitting the elastic tensor from deformation and stresses.

        Note, the elastic tensor are obtained by fitting to the Lagrangian strains
        and second Piola-Kirchhoff stresses. In continuum mechanics nomenclature,
        it's the `material elasticity tensor`. For more, see Section 6.5 of
        `Continuum Mechanics and Thermodynamics -- From Fundamental Concepts to
        Governing Equations` Tadmor, Miller, and Elliott, Cambridge University Press,
        2012.

        Args:
            structure: relaxed structure.
            material_id: material id.
            deformations: deformations applied to the relaxed structure that generate
                the strained structures.
            stresses: Cauchy stresses on the deformed structures. Expected units: GPa.
            deformation_task_ids: id of the deformation tasks.
            deformation_dir_names: directories where the deformation tasks are run.
            equilibrium_stress: Cauchy stress on the relaxed structure.
            optimization_task_id: id of the optimization task to relax the structure.
            optimization_dir_name: directory where the optimization task run.
            fitting_method: method used to fit the elastic tensor:
                {`finite_difference`, `pseudoinverse`, `independent`}.
        """
        CM = CriticalMessage()

        # primary fitting data
        p_deforms = deformations
        p_stresses = stresses
        (
            p_strains,
            p_2nd_pk_stresses,
            p_task_ids,
            p_dir_names,
        ) = generate_primary_fitting_data(
            p_deforms, p_stresses, deformation_task_ids, deformation_dir_names
        )

        # derived fitting data
        (
            d_deforms,
            d_strains,
            d_stresses,
            d_2nd_pk_stresses,
        ) = generate_derived_fitting_data(structure, p_strains, p_stresses)

        fitting_strains = p_strains + d_strains
        fitting_stresses = p_2nd_pk_stresses + d_2nd_pk_stresses

        # avoid symmop-related strains having non-symmop-related stresses
        fitting_stresses = symmetrize_stresses(
            fitting_stresses, fitting_strains, structure
        )

        # fitting elastic tensor
        try:
            elastic_tensor = fit_elastic_tensor(
                fitting_strains,
                fitting_stresses,
                eq_stress=equilibrium_stress,
                fitting_method=fitting_method,
            )

            # elastic and compliance tensors, only round ieee format ones
            ieee_et = elastic_tensor.voigt_symmetrized.convert_to_ieee(structure)
            et_doc = ElasticTensorDoc(
                raw=elastic_tensor.voigt.tolist(),
                ieee_format=ieee_et.round(0).voigt.tolist(),
            )

        except np.linalg.LinAlgError as e:
            et_doc = None
            ct_doc = None
            derived_props = {}
            state = TaskState("failed")
            warnings = [CM.FITTING.format(e)]

        else:
            try:
                compliance = elastic_tensor.compliance_tensor
                compliance_ieee = ieee_et.compliance_tensor

                # compliance tensor, *1000 to convert units to TPa^-1, i.e. 10^-12 Pa,
                # assuming elastic tensor in units of GPa
                ct_doc = ComplianceTensorDoc(
                    raw=(compliance * 1000).voigt.tolist(),
                    ieee_format=(compliance_ieee * 1000).round(0).voigt.tolist(),
                )

                # derived properties
                # (should put it here since some derived properties also dependent on
                # compliance tensor)
                derived_props = get_derived_properties(structure, elastic_tensor)

                # check all
                state, warnings = sanity_check(structure, et_doc, fitting_strains, derived_props)  # type: ignore

            except np.linalg.LinAlgError as e:
                ct_doc = None
                derived_props = {}
                state = TaskState("failed")
                warnings = [CM.COMPLIANCE.format(e)]

        # fitting data
        eq_stress = None if equilibrium_stress is None else equilibrium_stress.tolist()
        n_states = len(p_deforms) + len(d_deforms)
        if n_states != 24:
            warnings.append(CM.N_STATES.format(n_states))
            state = TaskState("failed")

        fitting_data = FittingData(
            deformations=[x.tolist() for x in p_deforms],  # type: ignore
            strains=[x.tolist() for x in p_strains],  # type: ignore
            cauchy_stresses=[x.tolist() for x in p_stresses],  # type: ignore
            second_pk_stresses=[x.tolist() for x in p_2nd_pk_stresses],  # type: ignore
            deformation_tasks=p_task_ids,  # type: ignore
            deformation_dir_names=p_dir_names,  # type: ignore
            equilibrium_cauchy_stress=eq_stress,
            optimization_task=optimization_task_id,
            optimization_dir_name=optimization_dir_name,  # type: ignore
            num_total_strain_stress_states=n_states,
        )

        return cls.from_structure(
            structure,
            material_id,
            order=2,
            elastic_tensor=et_doc,
            compliance_tensor=ct_doc,
            fitting_data=fitting_data,
            fitting_method=fitting_method,
            warnings=warnings,
            state=state,
            deprecated=state == TaskState("failed"),
            **derived_props,
            **kwargs,
        )


def generate_primary_fitting_data(
    deforms: list[Deformation],
    stresses: list[Stress],
    task_ids: list[IdentifierType] | None = None,
    dir_names: list[str] | None = None,
) -> tuple[list[Strain], list[Stress], list[IdentifierType] | None, list[str] | None]:
    """
    Get the primary fitting data, i.e. data obtained from a calculation.

    Args:
        deforms: primary deformations
        stresses: primary stresses
        task_ids: ids of the tasks that generate the data
        dir_names: directories where the calculations are performed

    Returns:
        strains: primary strains
        second_pk_stresses: primary second Piola-Kirchhoff stresses
        task_ids: ids of the tasks that generate the data
        dir_names: directories where the calculations are performed
    """
    size = len(deforms)
    assert len(stresses) == size
    if task_ids is not None:
        assert len(task_ids) == size
    if dir_names is not None:
        assert len(dir_names) == size

    strains = [d.green_lagrange_strain for d in deforms]
    second_pk_stresses = [s.piola_kirchoff_2(d) for (s, d) in zip(stresses, deforms)]

    return strains, second_pk_stresses, task_ids, dir_names


def generate_derived_fitting_data(
    structure: Structure,
    strains: list[Strain],
    stresses: list[Stress],
    symprec=SETTINGS.SYMPREC,
    tol: float = 0.002,
) -> tuple[list[Deformation], list[Strain], list[Stress], list[Stress]]:
    """
    Get the derived fitting data from symmetry operations on the primary fitting data.

    It can happen that multiple primary deformations can be mapped to the same derived
    deformation from different symmetry operations. Ideally, this cannot happen if one
    use the same structure to determine all the symmetry operations.

    However, this is not the case in atomate, where the deformation tasks are determined
    based on the symmetry of the structure before the tight relaxation, which in this
    function the structure is the relaxed structure. The symmetries can be different.

    In atomate2, this is not a problem, because the deformation tasks are determined
    based on the relaxed structure.

    To make it work for all cases, the stress for a derived deformation is the average
    of all derived stresses, each corresponding to a primary calculation.
    In doing so, we also check to ensure that:
    1. only independent derived deformations are used
    2. for a specific derived strain, a primary deformation is only used (mapped)
       once to obtain the average

    Args:
        structure: relaxed structure
        strains: primary strains
        stresses: primary stresses
        symprec: symmetry operation precision
        tol: tolerance for comparing strains and also for determining whether the
            deformation corresponds to the train is independent. The elastic workflow
            use a minimum strain of 0.005, so the default tolerance of 0.002 should be
            able to distinguish different strain states.

    Returns:
        derived_deforms: derived deformations
        derived_strains: derived strains
        derived_stresses: derived Cauchy stresses
        derived_2nd_pk_stresses: derived second Piola-Kirchhoff stresses
    """
    sga = SpacegroupAnalyzer(structure, symprec=symprec)
    symmops = sga.get_symmetry_operations(cartesian=True)

    # primary strain mapping (used only for checking purpose below)
    p_mapping = TensorMapping(strains, strains, tol=tol)

    # Warnings:
    # Do not use deformations to replace strains in generating the derived fitting
    # data. More specifically, do not create TensorMapping using deformation. This is
    # because the Lagrangian strain is symmetric, but the deformation gradient is not.
    # Then, more derived data can be generated than enough/necessary, due to the
    # asymmetry of the deformation gradient.

    # generated derived deforms
    mapping = TensorMapping(tol=tol)
    for i, p_strain in enumerate(strains):
        for op in symmops:
            d_strain = p_strain.transform(op)

            # sym op generates another primary strain
            if d_strain in p_mapping:
                continue

            # sym op generates a non-independent deform
            if not d_strain.get_deformation_matrix().is_independent(tol=tol):
                continue

            # seen this derived deform before
            if d_strain in mapping:
                # all the existing `i`
                current = [t[1] for t in mapping[d_strain]]

                if i not in current:
                    mapping[d_strain].append((op, i))

            # not seen this derived deform before
            else:
                mapping[d_strain] = [(op, i)]

    # get average stress from derived deforms
    derived_strains = []
    derived_stresses = []
    derived_deforms = []
    derived_2nd_pk_stresses = []

    for d_strain, op_set in mapping.items():
        symmops, p_indices = zip(*op_set)  # type: ignore[assignment]

        p_stresses = [stresses[i] for i in p_indices]
        d_stresses = [s.transform(op) for s, op in zip(p_stresses, symmops)]
        d_stress = Stress(np.average(d_stresses, axis=0))

        derived_strains.append(d_strain)
        derived_stresses.append(d_stress)

        deform = d_strain.get_deformation_matrix()
        derived_deforms.append(deform)
        derived_2nd_pk_stresses.append(d_stress.piola_kirchoff_2(deform))

    return derived_deforms, derived_strains, derived_stresses, derived_2nd_pk_stresses


def symmetrize_stresses(
    stresses: list[Stress],
    strains: list[Strain],
    structure: Structure,
    symprec=SETTINGS.SYMPREC,
    tol: float = 0.002,
) -> list[Stress]:
    """
    Symmetrize stresses by averaging over all symmetry operations.

    Args:
        stresses: stresses to be symmetrized
        strains: strains corresponding to the stresses
        structure: materials structure
        symprec: symmetry operation precision
        tol: tolerance for comparing strains and also for determining whether the
            deformation corresponds to the train is independent. The elastic workflow
            use a minimum strain of 0.005, so the default tolerance of 0.002 should be
            able to distinguish different strain states.

    Returns: symmetrized stresses
    """
    sga = SpacegroupAnalyzer(structure, symprec=symprec)
    symmops = sga.get_symmetry_operations(cartesian=True)

    # for each strain, get the stresses from other strain states related by symmetry
    symmmetrized_stresses = []  # type: list[Stress]
    for strain, _stress in zip(strains, stresses):
        mapping = TensorMapping([strain], [[]], tol=tol)
        for strain2, stress2 in zip(strains, stresses):
            for op in symmops:
                if strain2.transform(op) in mapping:
                    mapping[strain].append(stress2.transform(op))
        sym_stress = np.average(mapping[strain], axis=0)
        symmmetrized_stresses.append(Stress(sym_stress))

    return symmmetrized_stresses


def fit_elastic_tensor(
    strains: list[Strain],
    stresses: list[Stress],
    eq_stress: Stress | None,
    fitting_method: str = "finite_difference",
    order: int = 2,
) -> ElasticTensor:
    """
    Fitting the elastic tensor.

    Args:
        strains: strains (primary and derived) to fit the elastic tensor
        stresses: stresses (primary and derived) to fit the elastic tensor
        eq_stress: equilibrium stress, i.e. stress on the relaxed structure
        fitting_method: method used to fit the elastic tensor:
            {`finite_difference`, `pseudoinverse`, `independent`}
        order: expansion order of the elastic tensor, 2 or 3

    Returns:
        fitted elastic tensor
    """
    if order > 2 or fitting_method == "finite_difference":
        # force finite diff if order > 2
        result = ElasticTensorExpansion.from_diff_fit(
            strains, stresses, eq_stress=eq_stress, order=order
        )
        if order == 2:
            result: ElasticTensor = ElasticTensor(result[0])  # type: ignore[no-redef]
    elif fitting_method == "pseudoinverse":
        result: ElasticTensor = ElasticTensor.from_pseudoinverse(strains, stresses)  # type: ignore[no-redef]
    elif fitting_method == "independent":
        result: ElasticTensor = ElasticTensor.from_independent_strains(  # type: ignore[no-redef]
            strains, stresses, eq_stress=eq_stress
        )
    else:
        raise ValueError(f"Unsupported elastic fitting method {fitting_method}")

    return result  # type: ignore[return-value]


def get_derived_properties(
    structure: Structure, tensor: ElasticTensor
) -> dict[str, Any]:
    """
    Get derived elasticity properties.

    Args:
        structure: relaxed structure
        tensor: elastic tensor corresponding to the structure

    Returns:
        a dict of derived elasticity properties
    """
    try:
        prop_dict = tensor.get_structure_property_dict(structure)
        prop_dict.pop("structure")
        structure_prop_computed = True

    except ValueError:
        prop_dict = tensor.property_dict
        structure_prop_computed = False

    decimals = 3
    derived_prop = {
        "bulk_modulus": BulkModulus(
            voigt=np.round(prop_dict["k_voigt"], decimals),  # type: ignore[arg-type]
            reuss=np.round(prop_dict["k_reuss"], decimals),  # type: ignore[arg-type]
            vrh=np.round(prop_dict["k_vrh"], decimals),  # type: ignore[arg-type]
        ),
        "shear_modulus": ShearModulus(
            voigt=np.round(prop_dict["g_voigt"], decimals),  # type: ignore[arg-type]
            reuss=np.round(prop_dict["g_reuss"], decimals),  # type: ignore[arg-type]
            vrh=np.round(prop_dict["g_vrh"], decimals),  # type: ignore[arg-type]
        ),
        "young_modulus": np.round(prop_dict["y_mod"], 0),  # type: ignore[arg-type]
        "homogeneous_poisson": np.round(prop_dict["homogeneous_poisson"], decimals),  # type: ignore[arg-type]
        "universal_anisotropy": np.round(prop_dict["universal_anisotropy"], decimals),  # type: ignore[arg-type]
    }

    if structure_prop_computed:
        derived_prop.update(
            {
                "sound_velocity": SoundVelocity(
                    transverse=prop_dict["trans_v"],  # type: ignore[arg-type]
                    longitudinal=prop_dict["long_v"],  # type: ignore[arg-type]
                    snyder_acoustic=prop_dict["snyder_ac"],  # type: ignore[arg-type]
                    snyder_optical=prop_dict["snyder_opt"],  # type: ignore[arg-type]
                    snyder_total=prop_dict["snyder_total"],  # type: ignore[arg-type]
                ),
                "thermal_conductivity": ThermalConductivity(
                    clarke=prop_dict["clarke_thermalcond"],  # type: ignore[arg-type]
                    cahill=prop_dict["cahill_thermalcond"],  # type: ignore[arg-type]
                ),
                "debye_temperature": prop_dict["debye_temperature"],  # type: ignore[arg-type]
            }
        )

    return derived_prop


def sanity_check(
    structure: Structure,
    elastic_doc: ElasticTensorDoc,
    strains: list[Strain],
    derived_props: dict[str, Any],
) -> tuple[TaskState, list[str]]:
    """
    Post analysis to generate warnings if any.

    Returns:
        state: state of the calculation
        warnings: all warning messages. Messages starting with `Critical` are the
            ones resulting in a `failed` state.
    """
    failed = False
    warnings = []
    CM = CriticalMessage()
    WM = WarningMessage()

    # rank of all strains < 6?
    voigt_strains = [s.voigt for s in strains]
    rank = np.linalg.matrix_rank(voigt_strains)
    if rank != 6:
        failed = True
        warnings.append(CM.STRAIN_RANK.format(rank))

    # elastic tensor eigenvalues
    eig_vals, _ = np.linalg.eig(elastic_doc.raw)  # type: ignore
    if np.any(eig_vals < 0.0):
        warnings.append(WM.NEGATIVE_EIGVAL)

    # elastic tensor individual components
    et = np.asarray(elastic_doc.ieee_format)

    tol = 1e6
    if np.any(et > tol):
        warnings.append(WM.LARGE_COMPONENT.format(tol))

    # modulus
    low = 2.0
    high = 1000.0
    for p in ["bulk_modulus", "shear_modulus"]:
        doc = derived_props[p]
        doc = doc.model_dump()
        p = p.replace("_", " ")
        for name in ["voigt", "reuss", "vrh"]:
            v = doc[name]
            if v < 0:
                failed = True
                warnings.append(WM.NEGATIVE_MODULUS.format(name, p, v))
            elif v < low:
                warnings.append(WM.SMALL_MODULUS.format(name, p, v, low))
            elif v > high:
                warnings.append(WM.LARGE_MODULUS.format(name, p, v, high))

    # young's modulus (note it is in Pa, not GPa)
    high = 1e12
    v = derived_props["young_modulus"]
    if v > high:
        warnings.append(WM.LARGE_YOUNG_MODULUS.format(v, high))

    state = TaskState("failed") if failed else TaskState("successful")

    return state, warnings



"""Core definition of a Materials Document"""

from typing import Mapping

from pydantic import BaseModel, Field
from pymatgen.analysis.structure_analyzer import SpacegroupAnalyzer, oxide_type
from pymatgen.analysis.structure_matcher import StructureMatcher
from pymatgen.entries.computed_entries import ComputedStructureEntry

from emmet.core.base import EmmetMeta
from emmet.core.material import MaterialsDoc as CoreMaterialsDoc
from emmet.core.material import PropertyOrigin
from emmet.core.settings import EmmetSettings
from emmet.core.tasks import TaskDoc
from emmet.core.types.pymatgen_types.computed_entries_adapter import (
    ComputedStructureEntryType,
)
from emmet.core.types.typing import IdentifierType
from emmet.core.utils import utcnow
from emmet.core.vasp.calc_types import CalcType, RunType, TaskType

SETTINGS = EmmetSettings()


class BlessedCalcs(BaseModel, populate_by_name=True):
    GGA: ComputedStructureEntryType | None = Field(None)
    GGA_U: ComputedStructureEntryType | None = Field(None, alias="GGA+U")
    PBESol: ComputedStructureEntryType | None = Field(None, alias="PBEsol")
    SCAN: ComputedStructureEntryType | None = Field(None)
    R2SCAN: ComputedStructureEntryType | None = Field(None, alias="r2SCAN")
    HSE: ComputedStructureEntryType | None = Field(None)


class MaterialsDoc(CoreMaterialsDoc):
    calc_types: Mapping[IdentifierType, CalcType] | None = Field(  # type: ignore
        None,
        description="Calculation types for all the calculations that make up this material",
    )
    task_types: Mapping[IdentifierType, TaskType] | None = Field(
        None,
        description="Task types for all the calculations that make up this material",
    )
    run_types: Mapping[IdentifierType, RunType] | None = Field(
        None,
        description="Run types for all the calculations that make up this material",
    )

    origins: list[PropertyOrigin] | None = Field(
        None, description="Struct array for tracking the provenance of properties"
    )

    entries: BlessedCalcs | None = Field(
        None, description="Dictionary for tracking entries for VASP calculations"
    )

    @classmethod
    def from_tasks(
        cls,
        task_group: list[TaskDoc],
        structure_quality_scores: dict[
            str, int
        ] = SETTINGS.VASP_STRUCTURE_QUALITY_SCORES,
        use_statics: bool = SETTINGS.VASP_USE_STATICS,
        commercial_license: bool = True,
    ) -> "MaterialsDoc":
        """
        Converts a group of tasks into one material

        Args:
            task_group: List of task document
            structure_quality_scores: quality scores for various calculation types
            use_statics: Use statics to define a material
            commercial_license: Whether the data should be licensed with BY-C (otherwise BY-NC).
        """
        if len(task_group) == 0:
            raise Exception("Must have more than one task in the group.")

        # Metadata
        last_updated = max(task.last_updated for task in task_group)
        created_at = min(task.completed_at for task in task_group)
        task_ids = list({task.task_id for task in task_group})

        deprecated_tasks = {task.task_id for task in task_group if not task.is_valid}
        run_types = {task.task_id: task.run_type for task in task_group}
        task_types = {task.task_id: task.task_type for task in task_group}
        calc_types = {task.task_id: task.calc_type for task in task_group}

        structure_optimizations = [
            task for task in task_group if task.task_type == TaskType.Structure_Optimization  # type: ignore
        ]
        statics = [task for task in task_group if task.task_type == TaskType.Static]  # type: ignore
        structure_calcs = (
            structure_optimizations + statics
            if use_statics
            else structure_optimizations
        )

        validity_check = [doc for doc in structure_calcs if doc.is_valid]
        if not validity_check:
            raise ValueError("Group must contain at least one valid task")

        # Material ID
        possible_mat_ids = [task.task_id for task in structure_optimizations]

        if use_statics:
            possible_mat_ids += [task.task_id for task in statics]

        material_id = min(possible_mat_ids)

        # Always prefer a static over a structure opt
        structure_task_quality_scores = {"Structure Optimization": 1, "Static": 2}

        def _structure_eval(task: TaskDoc):
            """
            Helper function to order structures optimization and statics calcs by
            - Functional Type
            - Spin polarization
            - Special Tags
            - Energy
            """

            task_run_type = task.run_type
            _SPECIAL_TAGS = ["LASPH", "ISPIN"]
            special_tags = sum(
                (
                    task.input.parameters.get(tag, False)
                    if task.input.parameters
                    else False
                )
                for tag in _SPECIAL_TAGS
            )

            return (
                -1 * int(task.is_valid),
                -1 * structure_quality_scores.get(task_run_type.value, 0),
                -1 * structure_task_quality_scores.get(task.task_type.value, 0),
                -1 * special_tags,
                task.output.energy_per_atom,
            )

        best_structure_calc = sorted(structure_calcs, key=_structure_eval)[0]
        structure = best_structure_calc.output.structure

        # Initial Structures
        initial_structures = [task.input.structure for task in task_group]
        sm = StructureMatcher(
            ltol=0.1, stol=0.1, angle_tol=0.1, scale=False, attempt_supercell=False
        )
        initial_structures = [
            group[0] for group in sm.group_structures(initial_structures)
        ]

        # Deprecated
        deprecated = all(task.task_id in deprecated_tasks for task in structure_calcs)
        deprecated = deprecated or best_structure_calc.task_id in deprecated_tasks

        # Origins
        origins = [
            PropertyOrigin(
                name="structure",
                task_id=best_structure_calc.task_id,
                last_updated=best_structure_calc.last_updated,
            )
        ]

        # Entries
        # **current materials docs must contain at last one GGA or GGA+U entry

        # Always prefer a static over a structure opt
        entry_task_quality_scores = {"Structure Optimization": 1, "Static": 2}

        def _entry_eval(task: TaskDoc):
            """
            Helper function to order entries and statics calcs by
            - Spin polarization
            - Special Tags
            - Energy
            """

            _SPECIAL_TAGS = ["LASPH", "ISPIN"]
            special_tags = sum(
                (
                    task.input.parameters.get(tag, False)
                    if task.input.parameters
                    else False
                )
                for tag in _SPECIAL_TAGS
            )

            return (
                -1 * int(task.is_valid),
                -1 * entry_task_quality_scores.get(task.task_type.value, 0),
                -1 * special_tags,
                task.output.energy_per_atom,
            )

        # Entries
        # **current materials docs must contain at last one GGA or GGA+U entry
        entries = {}
        all_run_types = set(run_types.values())

        for rt in all_run_types:
            relevant_calcs = sorted(
                [doc for doc in structure_calcs if doc.run_type == rt and doc.is_valid],
                key=_entry_eval,
            )

            if relevant_calcs:
                best_task_doc = relevant_calcs[0]
                entry = ComputedStructureEntry(
                    composition=best_task_doc.output.structure.composition,
                    correction=0.0,
                    data={
                        "aspherical": best_task_doc.input.parameters.get(
                            "LASPH", False
                        ),
                        "last_updated": str(utcnow()),
                        "oxide_type": oxide_type(best_task_doc.output.structure),
                        "material_id": material_id,
                        "task_id": best_task_doc.task_id,
                    },
                    energy=best_task_doc.output.energy,
                    entry_id="{}-{}".format(material_id, rt.value),
                    parameters={
                        "hubbards": best_task_doc.input.hubbards,
                        "is_hubbard": best_task_doc.input.is_hubbard,
                        "potcar_spec": (
                            [dict(d) for d in best_task_doc.input.potcar_spec]
                            if best_task_doc.input.potcar_spec
                            else []
                        ),
                        "run_type": str(best_task_doc.run_type),
                    },
                    structure=best_task_doc.output.structure,
                )
                entries[rt] = entry

        if not any(
            run_type in entries
            for run_type in (RunType.GGA, RunType.GGA_U, RunType.r2SCAN)
        ):
            raise ValueError(
                "Individual material entry must contain at least one GGA, GGA+U, or r2SCAN calculation"
            )

        # Builder meta and license
        builder_meta = EmmetMeta(license="BY-C" if commercial_license else "BY-NC")

        return cls.from_structure(
            meta_structure=structure,
            material_id=material_id,
            last_updated=last_updated,
            created_at=created_at,
            task_ids=task_ids,
            calc_types=calc_types,
            run_types=run_types,
            task_types=task_types,
            initial_structures=initial_structures,
            deprecated=deprecated,
            deprecated_tasks=deprecated_tasks,
            origins=origins,
            entries=entries,
            builder_meta=builder_meta,
        )

    @classmethod
    def construct_deprecated_material(
        cls,
        task_group: list[TaskDoc],
        commercial_license: bool = True,
    ) -> "MaterialsDoc":
        """
        Converts a group of tasks into a deprecated material

        Args:
            task_group: List of task document
            commercial_license: Whether the data should be licensed with BY-C (otherwise BY-NC).
        """
        if len(task_group) == 0:
            raise Exception("Must have more than one task in the group.")

        # Metadata
        last_updated = max(task.last_updated for task in task_group)
        created_at = min(task.completed_at for task in task_group)
        task_ids = list({task.task_id for task in task_group})

        deprecated_tasks = {task.task_id for task in task_group}
        run_types = {task.task_id: task.run_type for task in task_group}
        task_types = {task.task_id: task.task_type for task in task_group}
        calc_types = {task.task_id: task.calc_type for task in task_group}

        # Material ID
        material_id = min([task.task_id for task in task_group])

        # Choose any random structure for metadata
        structure = SpacegroupAnalyzer(
            task_group[0].output.structure, symprec=0.1
        ).get_conventional_standard_structure()

        origins = [
            PropertyOrigin(
                name="structure",
                task_id=task_group[0].task_id,
                last_updated=task_group[0].last_updated,
            )
        ]

        # Deprecated
        deprecated = True

        # Builder meta and license
        builder_meta = EmmetMeta(license="BY-C" if commercial_license else "BY-NC")

        return cls.from_structure(
            meta_structure=structure,
            material_id=material_id,
            last_updated=last_updated,
            created_at=created_at,
            task_ids=task_ids,
            calc_types=calc_types,
            run_types=run_types,
            task_types=task_types,
            deprecated=deprecated,
            deprecated_tasks=deprecated_tasks,
            builder_meta=builder_meta,
            origins=origins,
        )



from typing import Annotated, TypeVar

from pydantic import BeforeValidator, WrapSerializer
from pymatgen.io.vasp.inputs import Poscar
from typing_extensions import TypedDict

from emmet.core.types.pymatgen_types.structure_adapter import (
    TypedStructureDict,
    pop_empty_structure_keys,
)

TypedPoscarDict = TypedDict(
    "TypedPoscarDict",
    {
        "@module": str,
        "@class": str,
        "structure": TypedStructureDict,
        "true_names": bool,
        "selective_dynamics": list[list[bool]],
        "velocities": list[list[float]],
        "predictor_corrector": list[list[float]],
        "comment": str,
    },
)


PoscarTypeVar = TypeVar("PoscarTypeVar", Poscar, TypedPoscarDict)


def pop_poscar_empty_structure_keys(poscar: PoscarTypeVar):
    if isinstance(poscar, dict):
        clean_structure = pop_empty_structure_keys(poscar["structure"])
        poscar["structure"] = clean_structure

        return Poscar.from_dict(poscar)  # type: ignore[arg-type]

    return poscar


PoscarType = Annotated[
    PoscarTypeVar,
    BeforeValidator(pop_poscar_empty_structure_keys),
    WrapSerializer(lambda x, nxt, info: x.as_dict(), return_type=TypedPoscarDict),
]



from pydantic import BaseModel, Field

__all__ = ["Component", "ExtractedMaterial"]


class Component(BaseModel):
    formula: str = Field(..., description="Formula of this component.")
    amount: str = Field(..., description="Amount of this component.")
    elements: dict[str, str] = Field(
        ..., description="Amount of each chemical elements in this component."
    )


class Values(BaseModel):
    values: list[float] | None = Field(None, description="list of values.")
    min_value: float | None = Field(None, description="Minimal value.")
    max_value: float | None = Field(None, description="Maximal value.")


class ExtractedMaterial(BaseModel):
    """
    Model for a material extracted from the literature
    """

    material_string: str = Field(
        ..., description="String of the material as written in paper."
    )
    material_formula: str = Field(
        ..., description="Normalized formula of the material."
    )
    material_name: str | None = Field(None, description="English name of the material.")

    phase: str | None = Field(
        None, description="Phase description of material, such as anatase."
    )
    is_acronym: bool | None = Field(
        None, description="Whether the material is an acronym, such as LMO for LiMn2O4."
    )

    composition: list[Component] = Field(
        ..., description="List of components in this material."
    )
    amounts_vars: dict[str, Values] = Field(
        {}, description="Amount variables (formula subscripts)."
    )
    elements_vars: dict[str, list[str]] = Field(
        {}, description="Chemical element variables"
    )

    additives: list[str] = Field([], description="list of additives, dopants, etc.")
    oxygen_deficiency: str | None = Field(
        None, description="Symbol indicating whether the materials is oxygen deficient."
    )



"""Core definition of a Q-Chem Task Document"""

# mypy: ignore-errors
from __future__ import annotations

import logging
import re
from collections import OrderedDict
from pathlib import Path
from typing import Any, TYPE_CHECKING

from custodian.qchem.jobs import QCJob
from monty.serialization import loadfn
from pydantic import BaseModel, Field
from pymatgen.core.structure import Molecule
from pymatgen.io.qchem.inputs import QCInput

from emmet.core.qchem.calc_types import CalcType, LevelOfTheory, TaskType
from emmet.core.qchem.calculation import Calculation, CalculationInput
from emmet.core.qchem.task import QChemStatus
from emmet.core.structure import MoleculeMetadata
from emmet.core.utils import arrow_incompatible

if TYPE_CHECKING:
    from typing_extensions import Self

__author__ = (
    "Evan Spotte-Smith <ewcspottesmith@lbl.gov>, Rishabh D. Guha <rdguha@lbl.gov>"
)

logger = logging.getLogger(__name__)
# _DERIVATIVE_FILES = ("GRAD", "HESS")


@arrow_incompatible
class OutputDoc(BaseModel):
    initial_molecule: Molecule = Field(None, description="Input Molecule object")
    optimized_molecule: Molecule | None = Field(
        None, description="Optimized Molecule object"
    )

    # TODO: Discuss with Evan if these go here
    # species_hash: str = Field(
    #     None,
    #     description="Weisfeiler Lehman (WL) graph hash using the atom species as the graph node attribute.",
    # )
    # coord_hash: str = Field(
    #     None,
    #     description="Weisfeiler Lehman (WL) graph hash using the atom coordinates as the graph node attribute.",
    # )

    # last_updated: datetime = Field(
    #     None,
    #     description = "Timestamp for the most recent calculation for this QChem task document",
    # )

    final_energy: float = Field(
        description="Final electronic energy for the calculation (units: Hartree)"
    )
    enthalpy: float | None = Field(
        None, description="Total enthalpy of the molecule (units: kcal/mol)"
    )
    entropy: float | None = Field(
        None, description="Total entropy of the molecule (units: cal/mol-K"
    )
    dipoles: dict[str, Any] | None = Field(
        None, description="Dipolar information from the output"
    )
    mulliken: list[Any] | None = Field(
        None, description="Mulliken atomic partial charges and partial spins"
    )
    resp: list[float | list[Any]] | None = Field(
        None,
        description="Restrained Electrostatic Potential (RESP) atomic partial charges",
    )
    nbo: dict[str, Any] | None = Field(
        None, description="Natural Bonding Orbital (NBO) output"
    )

    frequencies: dict[str, Any] | list | None = Field(
        None,
        description="The list of calculated frequencies if job type is freq (units: cm^-1)",
    )

    frequency_modes: list | str | None = Field(
        None,
        description="The list of calculated frequency mode vectors if job type is freq",
    )

    @classmethod
    def from_qchem_calc_doc(cls, calc_doc: Calculation) -> "OutputDoc":
        """
        Create a summary of QChem calculation outputs from a QChem calculation document.

        Parameters
        ----------
        calc_doc
            A QChem calculation document.
        kwargs
            Any other additional keyword arguments

        Returns
        --------
        OutputDoc
            The calculation output summary
        """
        return cls(
            initial_molecule=calc_doc.input.initial_molecule,
            optimized_molecule=calc_doc.output.optimized_molecule,
            # species_hash = self.species_hash, #The three entries post this needs to be checked again
            # coord_hash = self.coord_hash,
            # last_updated = self.last_updated,
            final_energy=calc_doc.output.final_energy,
            dipoles=calc_doc.output.dipoles,
            enthalpy=calc_doc.output.enthalpy,
            entropy=calc_doc.output.entropy,
            mulliken=calc_doc.output.mulliken,
            resp=calc_doc.output.resp,
            nbo=calc_doc.output.nbo_data,
            frequencies=calc_doc.output.frequencies,
            frequency_modes=calc_doc.output.frequency_modes,
        )


@arrow_incompatible
class InputDoc(BaseModel):
    initial_molecule: Molecule = Field(
        title="Input Structure",
        description="Input molecule and calc details for the QChem calculation",
    )

    prev_rem_params: dict[str, Any] | None = Field(
        None,
        description="Parameters from a previous qchem calculation in the series",
    )

    rem: dict[str, Any] = Field(
        None,
        description="Parameters from the rem section of the current QChem calculation",
    )

    level_of_theory: str | LevelOfTheory | None = Field(
        None, description="Level of theory used in the qchem calculation"
    )

    task_type: str | TaskType | None = Field(
        None,
        description="The type of the QChem calculation : optimization, single point ... etc.",
    )

    tags: list[str] | None = Field(
        None, title="tag", description="Metadata tagged to a given task."
    )

    solvation_lot_info: dict[str, Any] | str | None = Field(
        None,
        description="Str or Dict representation of the solvent method used for the calculation",
    )

    special_run_type: str | None = Field(
        None, description="Special workflow name (if applicable)"
    )

    smiles: str | None = Field(
        None,
        description="Simplified molecular-input line-entry system (SMILES) string for the molecule involved "
        "in this calculation.",
    )

    calc_type: str | CalcType | None = Field(
        None,
        description="A combined dictionary representation of the task type along with the level of theory used",
    )

    @classmethod
    def from_qchem_calc_doc(cls, calc_doc: Calculation) -> "InputDoc":
        """
        Create qchem calculation input summary from a qchem calculation document.

        Parameters
        ----------
        calc_doc
            A QChem calculation document.

        Returns
        --------
        InputDoc
            A summary of the input molecule and corresponding calculation parameters
        """
        try:
            lot_val = calc_doc.level_of_theory.value
        except AttributeError:
            lot_val = calc_doc.level_of_theory

        try:
            ct_val = calc_doc.calc_type.value
        except AttributeError:
            ct_val = calc_doc.calc_type
        # TODO : modify this to get the different variables from the task doc.
        return cls(
            initial_molecule=calc_doc.input.initial_molecule,
            rem=calc_doc.input.rem,
            level_of_theory=lot_val,
            task_type=calc_doc.task_type.value,
            tags=calc_doc.input.tags,
            solvation_lot_info=calc_doc.solvation_lot_info,
            # special_run_type = calc_doc.input.special_run_type,
            # smiles = calc_doc.input.smiles,
            calc_type=ct_val,
        )


@arrow_incompatible
class CustodianDoc(BaseModel):
    corrections: list[Any] | None = Field(
        None,
        title="Custodian Corrections",
        description="List of custodian correction data for calculation.",
    )

    job: dict[str, Any] | QCJob | None = Field(
        None,
        title="Custodian Job Data",
        description="Job data logged by custodian.",
    )


# AnalysisDoc? Is there a scope for AnalysisDoc in QChem?


@arrow_incompatible
class TaskDoc(MoleculeMetadata):
    """
    Calculation-level details about QChem calculations that would eventually take over the TaskDocument implementation
    """

    dir_name: str | Path | None = Field(
        None, description="The directory for this QChem task"
    )

    state: QChemStatus | None = Field(
        None, description="State of this QChem calculation"
    )

    calcs_reversed: list[Calculation] | None = Field(
        None,
        title="Calcs reversed data",
        description="Detailed data for each QChem calculation contributing to the task document.",
    )

    task_type: CalcType | TaskType | None = Field(
        None, description="the type of QChem calculation"
    )

    orig_inputs: CalculationInput | dict[str, Any] | None = Field(
        {}, description="Summary of the original Q-Chem inputs"
    )

    input: InputDoc | None = Field(
        None,
        description="The input molecule and calc parameters used to generate the current task document.",
    )

    output: OutputDoc | None = Field(
        None,
        description="The exact set of output parameters used to generate the current task document.",
    )

    # TODO: Implement entry dict

    custodian: list[CustodianDoc] | None = Field(
        None,
        title="Calcs reversed data",
        description="Detailed custodian data for each QChem calculation contributing to the task document.",
    )

    critic2: dict[str, Any] | None = Field(
        None, description="Outputs from the critic2 calculation if performed"
    )

    custom_smd: str | dict[str, Any] | None = Field(
        None,
        description="The seven solvent parameters necessary to define a custom_smd model",
    )

    additional_fields: dict[str, Any] | None = Field(
        None, description="Any miscellaneous fields passed to the pydantic model"
    )

    # TODO some sort of @validator s if necessary

    @classmethod
    def from_directory(
        cls,
        dir_name: Path | str,
        validate_lot: bool = True,
        store_additional_json: bool = True,
        additional_fields: dict[str, Any] = None,
        **qchem_calculation_kwargs,
    ) -> Self:
        """
        Create a task document from a directory containing QChem files.

        Parameters
        ----------
        dir_name
            The path to the folder containing the calculation outputs.
        validate_lot
            Flag for matching the basis and functional with the list of functionals consistent with MPCules.
            Defaults to True. Change to False if you want to create a TaskDoc with other basis sets and functionals.
        store_additional_json
            Whether to store additional json files in the calculation directory.
        additional_fields
            Dictionary of additional fields to add to output document.
        **qchem_calculation_kwargs
            Additional parsing options that will be passed to the
            :obj:`.Calculation.from_qchem_files` function.

        Returns
        -------
        QChemTaskDoc
            A task document for the calculation
        """
        logger.info(f"Getting task doc in: {dir_name}")

        additional_fields = additional_fields or {}
        dir_name = Path(dir_name)
        task_files = _find_qchem_files(dir_name)

        if len(task_files) == 0:
            raise FileNotFoundError("No QChem files found!")

        critic2 = {}
        custom_smd = {}
        calcs_reversed = []
        for task_name, files in task_files.items():
            if task_name == "orig":
                continue
            else:
                calc_doc = Calculation.from_qchem_files(
                    dir_name,
                    task_name,
                    **files,
                    **qchem_calculation_kwargs,
                    validate_lot=validate_lot,
                )
                calcs_reversed.append(calc_doc)
                # all_qchem_objects.append(qchem_objects)

        # Lists need to be reversed so that newest calc is the first calc, all_qchem_objects are also reversed to match
        calcs_reversed.reverse()

        # all_qchem_objects.reverse()
        custodian = _parse_custodian(dir_name)
        additional_json = None
        if store_additional_json:
            additional_json = _parse_additional_json(dir_name)
            for key, _ in additional_json.items():
                if key == "processed_critic2":
                    critic2["processed"] = additional_json["processed_critic2"]
                elif key == "cpreport":
                    critic2["cp"] = additional_json["cpreport"]
                elif key == "YT":
                    critic2["yt"] = additional_json["yt"]
                elif key == "bonding":
                    critic2["bonding"] = additional_json["bonding"]
                elif key == "solvent_data":
                    custom_smd = additional_json["solvent_data"]

        _orig_inputs = _parse_orig_inputs(dir_name)
        orig_inputs = (
            CalculationInput.from_qcinput(_orig_inputs) if _orig_inputs else {}
        )

        dir_name = get_uri(dir_name)  # convert to full path

        # only store objects from last calculation
        # TODO: If vasp implementation makes this an option, change here as well
        qchem_objects = None
        included_objects = None
        if qchem_objects:
            included_objects = list(qchem_objects.keys())

        # run_stats = _get_run_stats(calcs_reversed), Discuss whether this is something which is necessary in terms of QChem calcs
        doc = cls.from_molecule(
            meta_molecule=calcs_reversed[-1].input.initial_molecule,
            dir_name=dir_name,
            calcs_reversed=calcs_reversed,
            custodian=custodian,
            additional_json=additional_json,
            additional_fields=additional_fields,
            completed_at=calcs_reversed[0].completed_at,
            orig_inputs=orig_inputs,
            input=InputDoc.from_qchem_calc_doc(calcs_reversed[0]),
            output=OutputDoc.from_qchem_calc_doc(calcs_reversed[0]),
            state=_get_state(calcs_reversed),
            qchem_objects=qchem_objects,
            included_objects=included_objects,
            critic2=critic2,
            custom_smd=custom_smd,
            task_type=calcs_reversed[0].task_type,
        )

        # doc = doc.copy(update=additional_fields)
        doc = doc.model_copy(update=additional_fields)
        return doc

    @staticmethod
    def get_entry(
        calcs_reversed: list[Calculation], task_id: str | None = None
    ) -> dict:
        """
        Get a computed entry from a list of QChem calculation documents.

        Parameters
        ----------
        calcs_reversed
            A list of QChem calculation documents in reverse order.
        task_id
            The job identifier

        Returns
        --------
        Dict
            A dict of computed entries
        """

        entry_dict = {
            "entry_id": task_id,
            "task_id": task_id,
            "charge": calcs_reversed[0].output.molecule.charge,
            "spin_multiplicity": calcs_reversed[0].output.molecule.spin_multiplicity,
            "level_of_theory": calcs_reversed[-1].input.level_of_theory,
            "solvent": calcs_reversed[-1].input.solv_spec,
            "lot_solvent": calcs_reversed[-1].input.lot_solv_combo,
            "custom_smd": calcs_reversed[-1].input.custom_smd,
            "task_type": calcs_reversed[-1].input.task_spec,
            "calc_type": calcs_reversed[-1].input.calc_spec,
            "tags": calcs_reversed[-1].input.tags,
            "molecule": calcs_reversed[0].output.molecule,
            "composition": calcs_reversed[0].output.molecule.composition,
            "formula": calcs_reversed[
                0
            ].output.formula.composition.aplhabetical_formula,
            "energy": calcs_reversed[0].output.final_energy,
            "output": calcs_reversed[0].output.as_dict(),
            "critic2": calcs_reversed[
                0
            ].output.critic,  # TODO: Unclear about orig_inputs
            "last_updated": calcs_reversed[0].output.last_updated,
        }

        return entry_dict


def get_uri(dir_name: str | Path) -> str:
    """
    Return the URI path for a directory.

    This allows files hosted on different file servers to have distinct locations.

    Parameters
    ----------
    dir_name : str or Path
        A directory name.

    Returns
    -------
    str
        Full URI path, e.g., "fileserver.host.com:/full/payj/of/fir_name".
    """
    import socket

    fullpath = Path(dir_name).absolute()
    hostname = socket.gethostname()
    try:
        hostname = socket.gethostbyaddr(hostname)[0]
    except (socket.gaierror, socket.herror):
        pass
    return f"{hostname}:{fullpath}"


def _parse_custodian(dir_name: Path) -> dict | None:
    """
    Parse custodian.json file.

    Calculations done using custodian have a custodian.json file which tracks the makers
    performed and any errors detected and fixed.

    Parameters
    ----------
    dir_name
        Path to calculation directory.

    Returns
    --------
    dict | None
        The information parsed from custodian.json file.
    """
    filenames = tuple(dir_name.glob("custodian.json*"))
    if len(filenames) >= 1:
        return loadfn(filenames[0], cls=None)
    return None


def _parse_orig_inputs(
    dir_name: Path,
) -> dict[str, Any] | QCInput:
    """
    Parse original input files.

    Calculations using custodian generate a *.orig file for the inputs. This is useful
    to know how the calculation originally started.

    Parameters
    ----------
    dir_name
        Path to calculation directory.

    Returns
    -------
    dict[str, Any]
        The original molecule, rem, solvent and other data.
    """
    orig_inputs: dict[str, Any] | QCInput = {}
    if orig_file_path := next(dir_name.glob("*.orig*"), None):
        orig_inputs = QCInput.from_file(orig_file_path)
    return orig_inputs


def _parse_additional_json(dir_name: Path) -> dict[str, Any]:
    """Parse additional json files in the directory."""
    additional_json = {}
    for filename in dir_name.glob("*.json*"):
        key = filename.name.split(".")[0]
        if key not in ("custodian", "transformations"):
            if key not in additional_json:
                additional_json[key] = loadfn(filename, cls=None)
    return additional_json


def _get_state(calcs_reversed: list[Calculation]) -> QChemStatus:
    """Get state from calculation documents of QChem tasks."""
    all_calcs_completed = all(
        [c.has_qchem_completed == QChemStatus.SUCCESS for c in calcs_reversed]
    )
    if all_calcs_completed:
        return QChemStatus.SUCCESS
    return QChemStatus.FAILED


# def _get_run_stats(calcs_reversed: list[Calculation]) -> dict[str, RunStatistics]:
#     """Get summary of runtime statistics for each calculation in this task."""

#     run_stats = {}
#     total = dict(
#         average_memory=0.0,
#         max_memory=0.0,
#         elapsed_time=0.0,
#         system_time=0.0,
#         user_time=0.0,
#         total_time=0.0,
#         cores=0,
#     )


def _find_qchem_files(
    path: str | Path,
) -> dict[str, Any]:
    """
    Find QChem files in a directory.

    Only the mol.qout file (or alternatively files
    with the task name as an extension, e.g., mol.qout.opt_0.gz, mol.qout.freq_1.gz, or something like this...)
    will be returned.

    Parameters
    ----------
    path
        Path to a directory to search.

    Returns
    -------
    dict[str, Any]
        The filenames of the calculation outputs for each QChem task, given as a ordered dictionary of::

            {
                task_name:{
                    "qchem_out_file": qcrun_filename,
                },
                ...
            }
    If there is only 1 qout file task_name will be "standard" otherwise it will be the extension name like "opt_0"
    """
    path = Path(path)
    task_files = OrderedDict()

    in_file_pattern = re.compile(r"^(?P<in_task_name>mol\.(qin|in)(?:\..+)?)(\.gz)?$")

    for file in path.iterdir():
        if file.is_file():
            in_match = in_file_pattern.match(file.name)

            # This block is for generalizing outputs coming from both atomate and manual qchem calculations
            if in_match:
                in_task_name = re.sub(
                    r"(\.gz|gz)$",
                    "",
                    in_match.group("in_task_name").replace("mol.qin.", ""),
                )
                in_task_name = in_task_name or "mol.qin"
                if in_task_name == "orig":
                    task_files[in_task_name] = {"orig_input_file": file.name}
                elif in_task_name == "last":
                    continue
                elif in_task_name == "mol.qin" or in_task_name == "mol.in":
                    if in_task_name == "mol.qin":
                        out_file = (
                            path / "mol.qout.gz"
                            if (path / "mol.qout.gz").exists()
                            else path / "mol.qout"
                        )
                    else:
                        out_file = (
                            path / "mol.out.gz"
                            if (path / "mol.out.gz").exists()
                            else path / "mol.out"
                        )
                    task_files["standard"] = {
                        "qcinput_file": file.name,
                        "qcoutput_file": out_file.name,
                    }
                # This block will exist only if calcs were run through atomate
                else:
                    try:
                        task_files[in_task_name] = {
                            "qcinput_file": file.name,
                            "qcoutput_file": Path(
                                "mol.qout." + in_task_name + ".gz"
                            ).name,
                        }
                    except FileNotFoundError:
                        task_files[in_task_name] = {
                            "qcinput_file": file.name,
                            "qcoutput_file": "No qout files exist for this in file",
                        }

    return task_files



from typing import Annotated, TypeVar

from pydantic import BeforeValidator, WrapSerializer
from pymatgen.analysis.diffraction.xrd import DiffractionPattern
from typing_extensions import TypedDict


class TypedHKLDict(TypedDict):
    hkl: list[int]
    multiplicity: int


TypedDiffractionPatternDict = TypedDict(
    "TypedDiffractionPatternDict",
    {
        "@module": str,
        "@class": str,
        "x": list[float],
        "y": list[float],
        "hkls": list[list[TypedHKLDict]],
        "d_hkls": list[float],
    },
)

DiffractionPatternTypeVar = TypeVar(
    "DiffractionPatternTypeVar", DiffractionPattern, TypedDiffractionPatternDict
)

DiffractionPatternType = Annotated[
    DiffractionPatternTypeVar,
    BeforeValidator(
        lambda x: DiffractionPattern.from_dict(x) if isinstance(x, dict) else x
    ),
    WrapSerializer(
        lambda x, nxt, info: x.as_dict(), return_type=TypedDiffractionPatternDict
    ),
]



"""Define common testing utils used across emmet namespaces."""

from __future__ import annotations

import math
from contextlib import contextmanager
from importlib.resources import files as import_resource
from pathlib import Path
from tempfile import NamedTemporaryFile, TemporaryDirectory
from typing import TYPE_CHECKING

import numpy as np
import pandas as pd
from pydantic import BaseModel

from emmet.core.utils import arrow_incompatible

if TYPE_CHECKING:
    from collections.abc import Callable, Generator
    from typing import Any

    from emmet.core.types.typing import FSPathType


def _get_test_files_dir(module: str) -> Path:
    """Get path to test files directory."""
    return (
        Path(import_resource(module))  # type: ignore[arg-type]
        .parent.parent.parent.joinpath("test_files")
        .resolve()
    )


TEST_FILES_DIR = _get_test_files_dir("emmet.core")
"""Test files directory measured relative to emmet-core install."""


def _float_comparator(x: float, y: float):
    """Compare floats with defaults set to pytest.approx defaults."""
    return math.isclose(x, y, rel_tol=1e-6, abs_tol=1e-12)


def assert_schemas_equal(
    test_schema: Any, valid_schema: Any, float_comparator: Callable = _float_comparator
) -> None:
    """
    Recursively test all items in valid_schema are present and equal in test_schema.

    While test_schema can be a pydantic schema or dictionary, the valid schema must
    be a (nested) dictionary. This function automatically handles accessing the
    attributes of classes in the test_schema.

    Args:
        test_schema: A pydantic schema or dictionary of the schema.
        valid_schema: A (nested) dictionary specifying the key and values that must be
            present in test_schema.
        float_comparator (Callable) : A method to compare floats. Defaults to math.isclose.
            Should return a bool if two floats are approximately equal, and False otherwise.
    """

    if isinstance(valid_schema, dict):
        for key, sub_valid_schema in valid_schema.items():
            if isinstance(key, str) and hasattr(test_schema, key):
                sub_test_schema = getattr(test_schema, key, {})
                if key in ("initial_molecule", "optimized_molecule") and hasattr(
                    sub_test_schema, "as_dict"
                ):
                    sub_test_schema = sub_test_schema.as_dict()

            elif not isinstance(test_schema, BaseModel):
                sub_test_schema = test_schema[key]
            else:
                raise ValueError(f"{type(test_schema)} does not have field: {key}")
            return assert_schemas_equal(sub_test_schema, sub_valid_schema)

    elif isinstance(valid_schema, list):
        for i, sub_valid_schema in enumerate(valid_schema):
            return assert_schemas_equal(test_schema[i], sub_valid_schema)

    elif isinstance(valid_schema, np.ndarray):
        assert np.allclose(test_schema, valid_schema)

    elif isinstance(valid_schema, float):
        assert float_comparator(test_schema, valid_schema)
    else:
        assert test_schema == valid_schema


@arrow_incompatible
class DataArchive(BaseModel):
    """Tool to compress test data into a lower disk / innode use file."""

    files: dict[str, bytes]

    @classmethod
    def from_directory(cls, path: FSPathType):
        """Find all files in . and parse them."""

        files = {}
        for p in Path(path).glob("*"):
            if p.is_file():
                files[p.name] = p.read_bytes()
        return cls(files=files)

    def to_json(self, file_name: FSPathType) -> None:
        """Write all contents to JSON."""
        df = pd.DataFrame({k: [v.hex()] for k, v in self.files.items()})
        df.to_json(file_name, orient="columns")

    @staticmethod
    def from_json(archive_path: FSPathType) -> pd.DataFrame:
        df = pd.read_json(archive_path)
        for c in df.columns:
            df[c] = df[c].apply(bytes.fromhex)
        return df

    @classmethod
    @contextmanager  # type: ignore[arg-type]
    def extract(cls, archive_path: FSPathType) -> Generator[Path]:
        """Extract all bytes data from a JSON archive.

        Parameters
        -----------
        archive_path : FSPathType
            The name of the JSON file

        Returns
        -----------
        pandas DataFrame representing the parsed data.
        """

        df = cls.from_json(archive_path)
        temp_dir = TemporaryDirectory()
        base_path = Path(temp_dir.name)
        for c in df.columns:
            (base_path / c).write_bytes(df[c][0])

        try:
            yield base_path
        finally:
            temp_dir.cleanup()

    @staticmethod
    def extract_obj(
        archive_path: FSPathType, file_name: str, func: Callable, *args, **kwargs
    ) -> Any:
        """Extract a single file and apply a mapping to it."""
        df = pd.read_json(archive_path)
        if file_name not in df.columns:
            raise ValueError(
                f"No object {file_name} found, valid columns: {', '.join(df.columns)}"
            )

        temp_file = NamedTemporaryFile(suffix="".join(Path(file_name).suffixes))
        with open(temp_file.name, "wb") as f:
            f.write(bytes.fromhex(df[file_name][0]))
            f.seek(0)

        obj = func(temp_file.name, *args, **kwargs)
        temp_file.close()
        return obj



from typing import Annotated, TypeVar

from pydantic import BeforeValidator, WrapSerializer
from pymatgen.analysis.reaction_calculator import BalancedReaction
from typing_extensions import TypedDict

TypedBalancedReactionDict = TypedDict(
    "TypedBalancedReactionDict",
    {
        "@module": str,
        "@class": str,
        "reactants": dict[str, float],
        "products": dict[str, float],
    },
)

BalancedReactionTypeVar = TypeVar(
    "BalancedReactionTypeVar", BalancedReaction, TypedBalancedReactionDict
)

BalancedReactionType = Annotated[
    BalancedReactionTypeVar,
    BeforeValidator(
        lambda x: BalancedReaction.from_dict(x) if isinstance(x, dict) else x
    ),
    WrapSerializer(
        lambda x, nxt, info: x.as_dict(), return_type=TypedBalancedReactionDict
    ),
]



import pytest
from pymatgen.analysis.diffraction.xrd import WAVELENGTHS
from pymatgen.core import Element, Lattice, Structure

from emmet.core import ARROW_COMPATIBLE
from emmet.core.utils import jsanitize
from emmet.core.xrd import Edge, XRDDoc

if ARROW_COMPATIBLE:
    import pyarrow as pa

    from emmet.core.arrow import arrowize


@pytest.fixture
def structure():
    test_latt = Lattice.cubic(3.0)
    test_struc = Structure(lattice=test_latt, species=["Fe"], coords=[[0, 0, 0]])
    return test_struc


@pytest.mark.parametrize("target", list(WAVELENGTHS.keys()))
def test_target_detection(structure, target):
    doc = XRDDoc.from_structure(
        structure=structure,
        spectrum_id="test-1",
        material_id="test-1",
        wavelength=WAVELENGTHS[target],
    )

    target_element = Element(target[:2])
    target_edge = Edge(target[2:])
    assert doc.target == target_element
    assert doc.edge == target_edge


@pytest.mark.parametrize("target", list(WAVELENGTHS.keys()))
def test_from_target(structure, target):
    target_element = Element(target[:2])
    target_edge = Edge(target[2:])
    doc = XRDDoc.from_target(
        structure=structure,
        material_id="test-1",
        target=target_element,
        edge=target_edge,
    )
    assert doc.target == target_element
    assert doc.edge == target_edge


def test_schema():
    XRDDoc.model_json_schema()


@pytest.mark.skipif(
    not ARROW_COMPATIBLE, reason="pyarrow must be installed to run this test."
)
def test_arrow(structure):
    doc = XRDDoc.from_structure(
        structure=structure,
        spectrum_id="test-1",
        material_id="test-1",
        wavelength=WAVELENGTHS[next(iter(WAVELENGTHS))],
    )
    arrow_struct = pa.scalar(
        doc.model_dump(context={"format": "arrow"}), type=arrowize(XRDDoc)
    )
    test_arrow_doc = XRDDoc(**arrow_struct.as_py(maps_as_pydicts="strict"))

    assert jsanitize(doc.model_dump(), allow_bson=True) == jsanitize(
        test_arrow_doc.model_dump(), allow_bson=True
    )



from typing import Annotated, TypeVar

from pydantic import BeforeValidator, WrapSerializer
from pymatgen.analysis.xas.spectrum import XAS
from typing_extensions import TypedDict

from emmet.core.types.pymatgen_types.structure_adapter import (
    TypedStructureDict,
    pop_empty_structure_keys,
)

TypedElementDict = TypedDict(
    "TypedElementDict",
    {"@module": str, "@class": str, "element": str},
)

TypedXASSpectrumDict = TypedDict(
    "TypedXASSpectrumDict",
    {
        "@module": str,
        "@class": str,
        "@version": str,
        "x": list[float],
        "y": list[float],
        "structure": TypedStructureDict,
        "absorbing_element": TypedElementDict,
        "edge": str,
        "spectrum_type": str,
        "absorbing_index": int,
    },
)


XASTypeVar = TypeVar("XASTypeVar", XAS, TypedXASSpectrumDict)


def pop_xas_empty_structure_fields(xas: XASTypeVar):
    if isinstance(xas, dict):
        clean_structure = pop_empty_structure_keys(xas["structure"])
        xas["structure"] = clean_structure
        return XAS.from_dict(xas)

    return xas


XASType = Annotated[
    XASTypeVar,
    BeforeValidator(pop_xas_empty_structure_fields),
    WrapSerializer(lambda x, nxt, info: x.as_dict(), return_type=TypedXASSpectrumDict),
]



from pydantic import BaseModel, Field

from emmet.core.types.typing import DateTimeType
from emmet.core.utils import arrow_incompatible


@arrow_incompatible
class FermiDoc(BaseModel):
    """
    Fermi surfaces.
    """

    fermi_surfaces: list[dict] | None = Field(
        None,
        description="List of IFermi FermiSurface objects.",
    )

    surface_types: list[str] | None = Field(
        None,
        description="Type of each fermi surface in the fermi_surfaces list.\
            Is either CBM or VBM for semiconductors, or fermi_surface for metals.",
    )

    material_id: str | None = Field(
        None,
        description="The Materials Project ID of the material. This comes in the form: mp-******.",
    )

    last_updated: DateTimeType = Field(
        description="Timestamp for the most recent calculation for this fermi surface document.",
    )



import pytest

from emmet.core.qchem.task import TaskDocument
from emmet.core.molecules.orbitals import OrbitalDoc

from tests.conftest_qchem import safe_load


@pytest.fixture(scope="session")
def closed_shell(test_dir):
    task = TaskDocument(**safe_load((test_dir / "closed_shell_nbo_task.json.gz")))
    return task


@pytest.fixture(scope="session")
def open_shell(test_dir):
    task = TaskDocument(**safe_load((test_dir / "open_shell_nbo_task.json.gz")))
    return task


@pytest.fixture(scope="session")
def new_closed(test_dir):
    task = TaskDocument(**safe_load((test_dir / "closed_shell_with3c.json.gz")))
    return task


@pytest.fixture(scope="session")
def new_open(test_dir):
    task = TaskDocument(**safe_load((test_dir / "open_shell_with3c.json.gz")))
    return task


def test_orbital(closed_shell, open_shell):
    # Test closed-shell NBO parsing
    doc = OrbitalDoc.from_task(
        closed_shell, "b9ba54febc77d2a9177accf4605767db-C1Li2O3-1-2", deprecated=False
    )

    assert doc.property_name == "natural bonding orbitals"
    assert doc.open_shell is False

    assert len(doc.nbo_population) == len(closed_shell.output.initial_molecule)
    assert doc.nbo_population[0].valence_electrons == pytest.approx(2.75426)
    assert len(doc.nbo_lone_pairs) == 3
    assert doc.nbo_lone_pairs[0].s_character == pytest.approx(0.02)
    assert doc.nbo_lone_pairs[0].atom_index == 0
    assert len(doc.nbo_bonds) == 10
    assert doc.nbo_bonds[0].atom1_s_character == pytest.approx(29.93)
    assert doc.nbo_bonds[0].atom1_index == 0
    assert doc.nbo_bonds[0].atom2_index == 3
    assert len(doc.nbo_interactions) == 95
    assert doc.nbo_interactions[0].donor_index == 8
    assert doc.nbo_interactions[0].acceptor_index == 19
    assert doc.nbo_interactions[0].energy_difference == pytest.approx(0.95)
    assert doc.alpha_population is None
    assert doc.beta_population is None

    # Test open-shell NBO parsing
    doc = OrbitalDoc.from_task(
        open_shell, "b9ba54febc77d2a9177accf4605767db-C1Li2O3-1-2", deprecated=False
    )

    assert doc.open_shell is True

    assert len(doc.nbo_population) == len(open_shell.output.initial_molecule)
    assert doc.alpha_population is not None
    assert doc.beta_population is not None
    assert doc.nbo_lone_pairs is None
    assert doc.nbo_bonds is None


def test_new_parser(new_closed, new_open):
    # Test closed-shell parsing including 3-center bonds and hyperbonds
    cs_doc = OrbitalDoc.from_task(
        new_closed, "b9ba54febc77d2a9177accf4605767db-C1Li2O3-1-2", deprecated=False
    )

    assert cs_doc.open_shell is False
    assert len(cs_doc.nbo_three_center_bonds) == 3
    assert cs_doc.nbo_three_center_bonds[1].atom1_polarization == pytest.approx(34.47)
    assert cs_doc.nbo_hyperbonds[2].hybrid_index_2 == 184

    # Test open-shell parsing including 3-center bonds and hyperbonds
    os_doc = OrbitalDoc.from_task(
        new_open, "b9ba54febc77d2a9177accf4605767db-C1Li2O3-1-2", deprecated=False
    )

    assert os_doc.open_shell is True
    assert len(os_doc.alpha_three_center_bonds) == 3
    assert os_doc.alpha_three_center_bonds[0].atom3_index == 14
    assert len(os_doc.beta_hyperbonds) == 2
    assert os_doc.beta_hyperbonds[0].fraction_12 == pytest.approx(41.2)



"""Schemas and utils for NEB calculations."""

from __future__ import annotations

import numpy as np
from pathlib import Path
from typing import TYPE_CHECKING, Any

import orjson
from monty.os.path import zpath
from pydantic import (
    BaseModel,
    Field,
    field_serializer,
    field_validator,
    model_validator,
)
from pymatgen.core import Molecule, Structure
from scipy.interpolate import CubicSpline
from typing_extensions import Self

from emmet.core.tasks import (
    _VOLUMETRIC_FILES,
    CustodianDoc,
    InputDoc,
    OrigInputs,
    _find_vasp_files,
    _parse_custodian,
    _parse_orig_inputs,
)
from emmet.core.types.enums import ValueEnum, VaspObject
from emmet.core.types.pymatgen_types.structure_adapter import StructureType
from emmet.core.utils import arrow_incompatible, type_override
from emmet.core.types.typing import DateTimeType, NullableDateTimeType
from emmet.core.vasp.calculation import Calculation
from emmet.core.vasp.task_valid import TaskState

if TYPE_CHECKING:
    from collections.abc import Callable, Sequence

    from numpy.typing import NDArray


def _join_endpoint_and_image_data(
    endpoint_data: list[Any],
    image_data: list[Any] | None,
    map_fun: Callable | None = None,
) -> list[Any]:
    data = [endpoint_data[0], *(image_data or []), endpoint_data[1]]
    if map_fun:
        data = list(map(map_fun, data))
    return data


class NebMethod(ValueEnum):
    """Common methods for NEB calculations.

    TODO: convert to StrEnum
    """

    STANDARD = "standard"
    CLIMBING_IMAGE = "climbing_image"
    APPROX = "approx_neb"


class HopFailureReason(ValueEnum):
    """Define failure modes for ApproxNEB calculations."""

    ENDPOINT = "Endpoint structure relaxation failed."
    MIN_DIST = "Linear distance traversed by working ion is below threshold."
    MIN_IMAGE = "Too few image calculations succeeded."
    IMAGE_FAILURE = "No image calculation output."


class BarrierAnalysis(BaseModel):
    """Define analysis schema for barrier calculations."""

    energies: list[float] = Field(
        description="The energies of each frame along the reaction coordinate."
    )
    frame_index: list[float] | None = Field(
        None,
        description="The fractional index along the reaction coordinate, between 0 and 1.",
    )
    cubic_spline_pars: list[list[float]] | None = Field(
        None, description="Parameters of the cubic spline used to fit the energies."
    )
    ts_frame_index: float | None = Field(
        None, description="The fractional index of the reaction coordinate."
    )
    ts_energy: float | None = Field(
        None, description="The energy at the transition state."
    )
    ts_in_frames: bool | None = Field(
        None,
        description="Whether the transition state is one of the computed snapshots.",
    )
    forward_barrier: float | None = Field(None, description="The forward barrier.")
    reverse_barrier: float | None = Field(None, description="The reverse barrier.")

    @classmethod
    def from_energies(
        cls,
        energies: Sequence[float],
        frame_index: Sequence[float] | NDArray | None = None,
        spline_kwargs: dict[str, Any] | None = None,
        frame_match_tol: float = 1.0e-6,
    ) -> Self:
        """
        Define basic NEB analysis tools.

        Parameters
        ----------
        energies : Sequence of float
            The energies sorted by increasing frame index. Must include endpoints.
        frame_index : Sequence of float or None (default)
            If None, defaults to a linear interpolation between 0 and 1 for each
            energy in energies. If not None, specifies the indices of succcessful
            images.
        spline_kwargs : dict or None
            The kwargs to pass to the spline fit. Defaults to clamping the derivative
            to zero at the endpoints, consistent with the assumption that they
            represent minima along the potential energy surface.
        frame_match_tol : float = 1.e-6
            The tolerance for matching the transition state frame index to the
            input frame indices.
        """
        frame_index = frame_index or list(range(len(energies)))
        frame_index = np.array(frame_index) / max(frame_index)

        analysis: dict[str, Any] = {
            "energies": list(energies),
            "frame_index": list(frame_index),
        }
        energies = np.array(energies)  # type: ignore[assignment]

        spline_kwargs = spline_kwargs or {"bc_type": "clamped"}
        spline_fit = CubicSpline(frame_index, energies, **spline_kwargs)
        analysis["cubic_spline_pars"] = spline_fit.c.tolist()

        crit_points = spline_fit.derivative().roots(extrapolate=False)
        analysis["ts_frame_index"] = -1
        analysis["ts_energy"] = -np.inf
        for crit_point in crit_points:
            if (energy := spline_fit(crit_point)) > analysis[
                "ts_energy"
            ] and spline_fit(crit_point, 2) <= 0.0:
                analysis["ts_frame_index"] = crit_point
                analysis["ts_energy"] = float(energy)

        analysis["ts_in_frames"] = any(
            abs(analysis["ts_frame_index"] - idx)
            < frame_match_tol * max(idx, frame_match_tol)
            for idx in frame_index
        )
        analysis["forward_barrier"] = analysis["ts_energy"] - energies[0]
        analysis["reverse_barrier"] = analysis["ts_energy"] - energies[-1]

        return cls(**analysis)


@arrow_incompatible
class NebResult(BaseModel):
    """Container class to store high-level NEB calculation info.

    This is intended to be code-agnostic, whereas NebTaskDoc
    is VASP-specific.
    """

    images: list[Structure | Molecule] | None = Field(
        None,
        description=(
            "Structures/molecules along the reaction pathway, "
            "including endpoints, after NEB."
        ),
    )

    initial_images: list[Structure | Molecule] | None = Field(
        None,
        description="Structures/molecules along the reaction pathway, prior to NEB.",
    )

    image_indices: list[int] | None = Field(
        None,
        description="The indexes corresponding to initial_images of all successful image calculations.",
    )

    energies: list[float] | None = Field(
        None, description="Energies corresponding the structures in `images`."
    )

    state: TaskState | None = Field(
        None, description="Whether the NEB calculation succeeded."
    )

    neb_method: NebMethod | None = Field(
        None, description="The NEB method used for this calculation."
    )

    forward_barrier: float | None = Field(
        None,
        description=(
            "Forward barrier for this reaction, "
            "i.e., the transition state energy minus "
            "the reactant / initial configuration energy."
        ),
    )

    reverse_barrier: float | None = Field(
        None,
        description=(
            "Reverse barrier for this reaction, "
            "i.e., the transition state energy minus "
            "the product / final configuration energy."
        ),
    )

    barrier_analysis: BarrierAnalysis | None = Field(
        None, description="Analysis of the reaction barrier."
    )

    failure_reasons: list[HopFailureReason] | None = Field(
        None, description="Reasons why the barrier calculation failed."
    )

    tags: list[str] | None = Field(
        None, description="List of string metadata about the calculation."
    )

    identifier: str | None = Field(None, description="Identifier for the calculation.")

    dir_name: str | None = Field(
        None, description="Directory where calculation was run."
    )

    @model_validator(mode="after")
    def set_barriers(self) -> Self:
        """Perform analysis on barrier if needed."""
        if (
            (not self.forward_barrier or not self.reverse_barrier)
            and isinstance(self.energies, list)
            and len(self.energies) > 0
        ):
            self.barrier_analysis = BarrierAnalysis.from_energies(
                self.energies, frame_index=self.image_indices
            )
            for k in ("forward", "reverse"):
                setattr(
                    self,
                    f"{k}_barrier",
                    getattr(self.barrier_analysis, f"{k}_barrier", None),
                )
        return self

    @property
    def barrier_energy_range(self) -> float | None:
        """The maximum computed energy minus the minimum computed energy along the path."""
        if self.energies:
            return max(self.energies) - min(self.energies)
        return None

    @property
    def num_images(self) -> int | None:
        if self.images:
            return len(self.images)
        return None


@type_override({"objects": str})
class NebIntermediateImagesDoc(BaseModel):
    """Schema for high-level intermediate image NEB data."""

    energies: list[float] | None = Field(
        None, description="The final energies of the intermediate images."
    )
    images: list[StructureType] | None = Field(
        None, description="Final structures for each intermediate image."
    )

    initial_images: list[StructureType] | None = Field(
        None, description="The initial intermediate image structures."
    )

    calculations: list[Calculation] | None = Field(
        None, description="Full calculation output for the intermediate images."
    )

    dir_name: str | None = Field(
        None, description="Top-level NEB calculation directory."
    )

    directories: list[str] | None = Field(
        None, description="List of the directories where the NEB images are located."
    )

    state: TaskState | None = Field(
        None, description="Whether the NEB calculation succeeded."
    )
    neb_method: NebMethod | None = Field(
        None, description="The NEB method used for this calculation."
    )

    orig_inputs: OrigInputs | None = Field(
        None,
        description="The exact set of input parameters used to generate the current task document.",
    )

    objects: list[dict[VaspObject, Any]] | None = Field(
        None, description="VASP objects for each image calculation."
    )

    inputs: InputDoc | None = Field(
        None, description="Inputs used in this calculation."
    )

    custodian: list[CustodianDoc] | None = Field(
        None,
        description="Detailed custodian data for each VASP calculation contributing to the task document.",
    )

    completed_at: NullableDateTimeType = Field(
        description="Timestamp for when this task was completed"
    )

    task_label: str | None = Field(
        None, description="Label for the NEB calculation(s)."
    )

    @field_serializer("objects", mode="wrap")
    def objects_serializer(self, d, default_serializer, info):
        default_serialized_object = default_serializer(d, info)

        format = info.context.get("format") if info.context else None
        if format == "arrow":
            return orjson.dumps(default_serialized_object)

        return default_serialized_object

    @field_validator("objects", mode="before")
    def objects_deserializer(cls, d):
        return orjson.loads(d) if isinstance(d, str) else d

    @classmethod
    def from_directory(
        cls,
        dir_name: str | Path,
        volumetric_files: Sequence[str] = _VOLUMETRIC_FILES,
        store_calculations: bool = True,
        **kwargs,
    ) -> Self:
        """
        Return an NebTaskDoc from a single NEB calculation directory.

        This method populates only the image energies and calculations fields,
        and the endpoint structures.
        """
        if isinstance(dir_name, str):
            dir_name = Path(dir_name)

        neb_directories = sorted(dir_name.glob("[0-9][0-9]"))

        image_directories = neb_directories[1:-1]

        image_calculations = []
        initial_structures = []
        image_structures = []
        image_objects = []
        for iimage, image_dir in enumerate(image_directories):
            vasp_files = _find_vasp_files(image_dir, volumetric_files=volumetric_files)

            calc, objects = Calculation.from_vasp_files(
                dir_name=image_dir,
                task_name=f"NEB image {iimage + 1}",
                vasprun_file=vasp_files["standard"]["vasprun_file"],
                outcar_file=vasp_files["standard"]["outcar_file"],
                contcar_file=vasp_files["standard"]["contcar_file"],
                volumetric_files=vasp_files["standard"].get("volumetric_files", []),
                oszicar_file=vasp_files["standard"].get("oszicar_file", None),
                vasprun_kwargs={
                    "parse_potcar_file": False,
                },
            )
            image_calculations.append(calc)
            image_structures.append(calc.output.structure)
            initial_structures.append(calc.input.structure)
            image_objects.append(objects)

        task_state = (
            TaskState.SUCCESS
            if all(
                calc.has_vasp_completed == TaskState.SUCCESS
                for calc in image_calculations
            )
            else TaskState.FAILED
        )

        inputs = {}
        for suffix in (None, ".orig"):
            vis = {
                k.lower(): v
                for k, v in _parse_orig_inputs(dir_name, suffix=suffix).items()
            }
            if (potcar_spec := vis.get("potcar")) is not None:
                vis["potcar_spec"] = potcar_spec
                vis["potcar"] = [spec.titel for spec in potcar_spec]

            if suffix is None:
                inputs["inputs"] = InputDoc(
                    **vis, magnetic_moments=vis.get("incar", {}).get("MAGMOM")
                )
            else:
                inputs["orig_inputs"] = OrigInputs(**vis)

        neb_method = (
            NebMethod.CLIMBING_IMAGE
            if inputs["inputs"].incar.get("LCLIMB", False)
            else NebMethod.STANDARD
        )

        return cls(
            calculations=image_calculations if store_calculations else None,
            images=image_structures,
            initial_images=initial_structures,
            dir_name=str(dir_name),
            directories=[str(img_dir) for img_dir in image_directories],
            orig_inputs=inputs["orig_inputs"],
            inputs=inputs["inputs"],
            objects=image_objects,
            neb_method=neb_method,  # type: ignore[arg-type]
            state=task_state,
            energies=[calc.output.energy for calc in image_calculations],
            custodian=_parse_custodian(dir_name),
            completed_at=max(calc.completed_at for calc in image_calculations),
            **kwargs,
        )


class NebTaskDoc(NebResult):
    """Define schema for VASP NEB tasks."""

    images: list[StructureType] | None = Field(  # type: ignore[assignment]
        None,
        description=(
            "Structures (including endpoints) along the reaction pathway after NEB."
        ),
    )

    initial_images: list[StructureType] | None = Field(  # type: ignore[assignment]
        None,
        description="Structures (including endpoints) along the reaction pathway prior to NEB.",
    )

    directories: list[str] | None = Field(
        None, description="Calculation directories for each image."
    )

    calculations: list[Calculation] | None = Field(
        None, description="The VASP calculations associated with each image."
    )

    last_updated: DateTimeType = Field(
        description="Timestamp for the most recent calculation for this task document",
    )

    objects: list[dict[VaspObject, Any] | None] | None = Field(
        None, description="VASP objects associated with each image calculation."
    )

    @classmethod
    def from_directories(
        cls,
        endpoint_directories: list[str | Path],
        neb_directory: str | Path,
        volumetric_files: Sequence[str] = _VOLUMETRIC_FILES,
        store_calculations: bool = True,
        **kwargs,
    ) -> Self:
        """
        Return an NebTaskDoc from endpoint and NEB calculation directories.

        This method populates the endpoint and image fields completely,
        permitting an analysis of the barrier.
        """

        ep_dirs = [Path(ep_dir).resolve() for ep_dir in endpoint_directories]
        neb_dir = Path(neb_directory).resolve()

        endpoint_calculations: list[Calculation | None] = [None, None]
        endpoint_objects = [None, None]
        for idx, endpoint_dir in enumerate(ep_dirs):
            vasp_files = _find_vasp_files(
                endpoint_dir, volumetric_files=volumetric_files
            )

            if vasp_files.get("standard"):
                ep_key = "standard"
            else:
                max_rel_idx = max(
                    int(k.split("relax")[-1])
                    for k in vasp_files
                    if k.startswith("relax")
                )
                ep_key = f"relax{max_rel_idx}"

            (
                endpoint_calculations[idx],
                endpoint_objects[idx],
            ) = Calculation.from_vasp_files(
                dir_name=endpoint_dir,
                task_name=f"NEB endpoint {idx + 1}",
                vasprun_file=vasp_files[ep_key]["vasprun_file"],
                outcar_file=vasp_files[ep_key]["outcar_file"],
                contcar_file=vasp_files[ep_key]["contcar_file"],
                volumetric_files=vasp_files[ep_key].get("volumetric_files", []),
                oszicar_file=vasp_files[ep_key].get("oszicar_file", None),
                vasprun_kwargs={
                    "parse_potcar_file": False,
                },
            )

        img_dirs = sorted(neb_dir.glob("[0-9][0-9]"))
        ep_img_dirs = [img_dirs[0], img_dirs[-1]]
        endpoint_energies: list[float | None] = [None, None]
        endpoint_structures: list[Structure | None] = [None, None]
        for idx, ep_calc in enumerate(endpoint_calculations):
            if ep_calc is None:
                ep_dirs[idx] = ep_img_dirs[idx]
                endpoint_structures[idx] = Structure.from_file(
                    zpath(f"{ep_dirs[idx]}/POSCAR")
                )
            else:
                endpoint_structures[idx] = ep_calc.output.structure
                endpoint_energies[idx] = ep_calc.output.energy

        intermediate_images = NebIntermediateImagesDoc.from_directory(
            neb_dir,
            volumetric_files=volumetric_files,
            store_calculations=store_calculations,
            **kwargs,
        )

        states = [
            getattr(ep_calc, "state", TaskState.FAILED)
            for ep_calc in endpoint_calculations
        ] + [intermediate_images.state]

        state = TaskState.SUCCESS
        if any(calc_state == TaskState.FAILED for calc_state in states):
            state = TaskState.FAILED
        elif any(calc_state == TaskState.ERROR for calc_state in states):
            state = TaskState.ERROR

        images = _join_endpoint_and_image_data(
            endpoint_structures, intermediate_images.images
        )
        calculations = None
        if store_calculations:
            calculations = _join_endpoint_and_image_data(
                endpoint_calculations, intermediate_images.calculations
            )

        return cls(
            images=images,
            initial_images=_join_endpoint_and_image_data(
                endpoint_structures, intermediate_images.initial_images
            ),
            image_indices=list(range(len(images))),
            energies=_join_endpoint_and_image_data(
                endpoint_energies, intermediate_images.energies
            ),
            state=state,
            neb_method=intermediate_images.neb_method,
            dir_name=str(intermediate_images.dir_name),
            directories=_join_endpoint_and_image_data(
                ep_dirs, intermediate_images.directories, map_fun=str
            ),
            calculations=calculations,
            objects=_join_endpoint_and_image_data(
                endpoint_objects, intermediate_images.objects
            ),
            **kwargs,
        )


@arrow_incompatible
class NebPathwayResult(BaseModel):  # type: ignore[call-arg]
    """Class for containing multiple NEB calculations, as along a reaction pathway."""

    hops: dict[str, NebResult] = Field(
        description="Dict of NEB calculations included in this calculation"
    )

    forward_barriers: dict[str, float | None] | None = Field(
        None, description="Dict of the forward barriers computed here."
    )

    reverse_barriers: dict[str, float | None] | None = Field(
        None, description="Dict of the reverse barriers computed here."
    )

    identifier: str | None = Field(None, description="Identifier for the calculation.")

    tags: list[str] | None = Field(
        None, description="List of string metadata about the calculation."
    )

    host_structure: StructureType | None = Field(
        None, description="The structure without active/mobile site(s)."
    )

    host_formula: str | None = Field(
        None,
        description="The chemical formula of the structure without active site(s).",
    )

    host_formula_reduced: str | None = Field(
        None,
        description="The reduced chemical formula of the structure without active site(s).",
    )

    host_chemsys: str | None = Field(
        None,
        description="The chemical system for the structure without active site(s).",
    )

    active_species: str | None = Field(
        None, description="The formula of the active/mobile species."
    )

    @model_validator(mode="after")
    def set_top_levels(self) -> Self:
        """Set barriers and host structure metadata, if needed."""
        for direction in ("forward", "reverse"):
            if getattr(self, f"{direction}_barriers", None) is None:
                setattr(
                    self,
                    f"{direction}_barriers",
                    {
                        idx: getattr(neb_calc, f"{direction}_barrier", None)
                        for idx, neb_calc in self.hops.items()
                    },
                )

        if self.host_structure is not None:
            self.host_formula = self.host_structure.formula
            self.host_formula_reduced = self.host_structure.reduced_formula
            self.host_chemsys = self.host_structure.chemical_system

        return self

    @property
    def max_barriers(self) -> dict[str, float | None] | None:
        """Retrieve the maximum barrier along each hop."""

        barriers: dict[str, list[float]] = {}
        for b in [self.forward_barriers, self.reverse_barriers]:
            if b:
                for idx, barrier in b.items():
                    if barrier:
                        if idx not in barriers:
                            barriers[idx] = []
                        barriers[idx].append(barrier)

        if len(barriers) == 0:
            return None

        return {idx: max(v) for idx, v in barriers.items()}

    @property
    def barrier_ranges(self) -> dict[str, float | None]:
        """Retrieve the max minus min computed energy along each hop."""
        return {
            idx: neb_calc.barrier_energy_range for idx, neb_calc in self.hops.items()
        }



from typing import Annotated, Any, TypeVar

import orjson
from pydantic import BeforeValidator, WrapSerializer
from pymatgen.entries.computed_entries import ComputedEntry, ComputedStructureEntry
from typing_extensions import NotRequired, TypedDict

from emmet.core.types.pymatgen_types.element_adapter import ElementType
from emmet.core.types.pymatgen_types.structure_adapter import (
    TypedStructureDict,
    pop_empty_structure_keys,
)
from emmet.core.vasp.calculation import PotcarSpec

# TypedEnergyAdjustmentDict = TypedDict(
#     "TypedEnergyAdjustmentDict",
#     {
#         "@module": str,
#         "@class": str,
#         "@version": str,
#         "value": float,
#         "uncertainty": float,
#         "name": str,
#         "cls": MaterialsProject2020Compatibility | MaterialsProjectDFTMixingScheme,
#         "description": str,
#     },
# )

# TypedCompositionEnergyAdjustmentDict = TypedDict(
#     "TypedCompositionEnergyAdjustmentDict",
#     {
#         "@module": str,
#         "@class": str,
#         "@version": str,
#         "adj_per_atom": float,
#         "n_atoms": int,
#         "uncertainty_per_atom": float,
#         "name": str,
#         "cls": MaterialsProject2020Compatibility | MaterialsProjectDFTMixingScheme,
#         "description": str,
#     },
# )

# TypedTemperatureEnergyAdjustmentDict = TypedDict(
#     "TypedTemperatureEnergyAdjustmentDict",
#     {
#         "@module": str,
#         "@class": str,
#         "@version": str,
#         "adj_per_deg": float,
#         "temp": float,
#         "n_atoms": int,
#         "uncertainty_per_deg": float,
#         "name": str,
#         "cls": MaterialsProject2020Compatibility | MaterialsProjectDFTMixingScheme,
#         "description": str,
#     },
# )


class TypedCEDataDict(TypedDict):
    oxide_type: str
    aspherical: bool
    last_updated: str
    task_id: str
    material_id: str
    oxidation_states: dict[ElementType, float]
    license: str
    run_type: str


class TypedCEParameterDict(TypedDict):
    potcar_spec: list[PotcarSpec]
    run_type: str
    is_hubbard: bool
    hubbards: NotRequired[dict[str, float] | None]  # type: ignore[type-arg]


TypedComputedEntryDict = TypedDict(
    "TypedComputedEntryDict",
    {
        "@module": str,
        "@class": str,
        "energy": float,
        "composition": dict[ElementType, float],
        "entry_id": str,
        "correction": float,
        # "energy_adjustments": list[
        #     TypedCompositionEnergyAdjustmentDict
        #     | TypedEnergyAdjustmentDict
        #     | TypedTemperatureEnergyAdjustmentDict
        # ],
        "energy_adjustments": str,
        "parameters": TypedCEParameterDict,
        "data": TypedCEDataDict,
    },
)


class TypedComputedStructureEntryDict(TypedComputedEntryDict):
    structure: TypedStructureDict


ComputedEntryTypeVar = TypeVar(
    "ComputedEntryTypeVar",
    ComputedEntry,
    TypedComputedEntryDict,
)


ComputedStructureEntryTypeVar = TypeVar(
    "ComputedStructureEntryTypeVar",
    ComputedStructureEntry,
    TypedComputedStructureEntryDict,
)


def entry_serializer(entry, nxt, info) -> dict[str, Any]:
    default_serialized_object = nxt(entry.as_dict(), info)

    format = info.context.get("format") if info.context else None
    if format == "arrow":
        default_serialized_object["energy_adjustments"] = orjson.dumps(
            default_serialized_object["energy_adjustments"]
        )

    return default_serialized_object


def pop_cse_empty_keys(cse: dict) -> dict[str, Any]:
    if cse.get("structure"):
        cse["structure"] = pop_empty_structure_keys(cse["structure"])
    cse["data"] = {k: v for k, v in cse["data"].items() if v}  # type: ignore[typeddict-item]
    cse["parameters"] = {k: v for k, v in cse["parameters"].items() if v}  # type: ignore[typeddict-item]

    return cse


def entry_deserializer(entry: dict[str, Any] | ComputedEntry | ComputedStructureEntry):
    if isinstance(entry, dict):
        entry_cls: type[ComputedEntry | ComputedStructureEntry]
        match entry["@class"]:
            case "ComputedEntry":
                entry_cls = ComputedEntry
            case "ComputedStructureEntry":
                entry_cls = ComputedStructureEntry
                entry = pop_cse_empty_keys(entry)

        if isinstance(entry["energy_adjustments"], str):
            entry["energy_adjustments"] = orjson.loads(entry["energy_adjustments"])

        return entry_cls.from_dict(entry)  # type: ignore[arg-type]

    return entry


ComputedEntryType = Annotated[
    ComputedEntryTypeVar,
    BeforeValidator(entry_deserializer),
    WrapSerializer(entry_serializer, return_type=dict[str, Any]),
]


ComputedStructureEntryType = Annotated[
    ComputedStructureEntryTypeVar,
    BeforeValidator(entry_deserializer),
    WrapSerializer(entry_serializer, return_type=dict[str, Any]),
]



"""Define utilities needed for parsing VASP calculations."""

from __future__ import annotations

import logging
import os
from collections import defaultdict
from pathlib import Path
from typing import TYPE_CHECKING

from pydantic import BaseModel, Field, model_validator

from emmet.core.utils import get_hash_blocked

if TYPE_CHECKING:
    from typing import Any

    from emmet.core.types.typing import FSPathType

logger = logging.getLogger(__name__)

TASK_NAMES = {"precondition"} | {f"relax{i+1}" for i in range(9)}


class FileMetadata(BaseModel):
    """
    Lightweight model to enable validation on files via MD5.

    """

    name: str = Field(
        description="Name of the VASP file without suffixes (e.g., INCAR)"
    )
    path: Path = Field(description="Path to the VASP file")
    hash: str | None = Field(
        description="Hash of the file (computed only when requested)",
        default=None,
    )

    @model_validator(mode="before")
    def coerce_path(cls, v: Any) -> Any:
        """Only coerce to Path. No existence check here."""
        if "path" in v:
            path = v["path"]
            if not isinstance(path, Path):
                path = Path(path)
            v["path"] = path
        return v

    def compute_hash(self) -> str | None:
        """Compute the hash of the file."""
        if self.validate_path_exists():
            try:
                self.hash = get_hash_blocked(self.path)
            except Exception:
                self.hash = None

        return self.hash

    def validate_path_exists(self):
        if not self.path.exists():
            raise ValueError(f"Path does not exist: {self.path}")
        if not self.path.is_file():
            raise ValueError(f"Path is not a file: {self.path}")
        return True

    def __hash__(self):
        return hash(self.path)

    def __eq__(self, other):
        if not isinstance(other, FileMetadata):
            return NotImplemented
        return self.path == other.path

    @property
    def calc_suffix(self) -> str:
        """Get any calculation-related suffixes, e.g., relax1."""
        suffix: str | None = None
        suffixes = self.path.suffixes
        for i in range(len(suffixes)):
            if (s := suffixes[-1 - i].split(".")[-1]) in TASK_NAMES:
                suffix = s
                break
        return suffix or "standard"


class CalculationLocator(BaseModel):
    """Object to represent calculation directory with possible file suffixes."""

    path: Path = Field(description="The path to the calculation directory")
    modifier: str | None = Field(
        description="Optional modifier for the calculation", default=None
    )

    def __hash__(self) -> int:
        # Resolve path to handle different representations of same path
        return hash((self.path.resolve(), self.modifier))

    # You might not need custom __eq__ in Pydantic v2
    # But if you keep it, consider path resolution:
    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, CalculationLocator):
            return False
        return (
            self.path.resolve() == other.path.resolve()
            and self.modifier == other.modifier
        )


VASP_INPUT_FILES = [
    "INCAR",
    "KPOINTS",
    "KPOINTS_OPT",
    "POSCAR",
    "POTCAR",
    "POTCAR.spec",
    "vaspin.h5",
]
VASP_ELECTRONIC_STRUCTURE = [
    "EIGENVAL",
    "DOSCAR",
]
VASP_VOLUMETRIC_FILES = (
    [
        "CHGCAR",
        "CHG",
    ]
    + [f"AECCAR{i}" for i in range(3)]
    + [
        "ELFCAR",
        "LOCPOT",
        "POT",
    ]
)
VASP_OUTPUT_FILES = [
    "CONTCAR",
    "IBZKPT",
    "OSZICAR",
    "OUTCAR",
    "PCDAT",
    "PROCAR",
    "REPORT",
    "vasprun.xml",
    "vaspout.h5",
    "XDATCAR",
]

VASP_RAW_DATA_ORG = {
    "input": VASP_INPUT_FILES.copy(),
    "output": VASP_OUTPUT_FILES.copy(),
    "volumetric": VASP_VOLUMETRIC_FILES.copy(),
    "electronic_structure": VASP_ELECTRONIC_STRUCTURE.copy(),
    "workflow": ["FW.json", "custodian.json", "transformations.json"],
}

REQUIRED_VASP_FILES = {"INCAR", "POSCAR", "POTCAR", "CONTCAR", "OUTCAR", "vasprun.xml"}

_vasp_files = set()
for v in VASP_RAW_DATA_ORG.values():
    _vasp_files.update(v)

VASP_RAW_DATA_ORG["input"].extend([f"{f}.orig" for f in VASP_INPUT_FILES])


def discover_vasp_files(
    target_dir: FSPathType,
) -> dict[str, list[FileMetadata]]:
    """
    Scan a target directory and identify VASP files.

    Parameters
    -----------
    target_dir : FSPathType

    Returns
    -----------
    List of FileMetadata for the identified files.
    """

    head_dir = Path(target_dir)
    vasp_files: list[FileMetadata] = []
    with os.scandir(head_dir) as scan_dir:
        for p in scan_dir:
            # Check that at least one VASP file matches the file name
            if p.is_file() and any(f for f in _vasp_files if f in p.name):
                vasp_files.append(FileMetadata(name=p.name, path=Path(p.path)))
    by_suffix = defaultdict(list)
    for file_meta in vasp_files:
        by_suffix[file_meta.calc_suffix].append(file_meta)
    return dict(by_suffix)  # type: ignore[arg-type]


def discover_and_sort_vasp_files(
    target_dir: FSPathType,
) -> dict[str, dict[str, Path | list[Path]]]:
    """
    Find and sort VASP files from a directory for TaskDoc.

    Parameters
    -----------
    target_dir : FSPathType

    Returns
    -----------
    dict of str (categories) to list of FileMetadata (list of VASP files in
        that category)
    """
    by_type: defaultdict[str, dict[str, Path | list[Path]]] = defaultdict(dict)
    for calc_suffix, files in discover_vasp_files(target_dir).items():
        for _f in files:
            f = _f.name.lower()
            file_path = _f.path.resolve()

            for k in ("vasprun", "contcar", "outcar", "potcar.spec"):
                if k in f:
                    by_type[calc_suffix][f"{k.replace('.','_')}_file"] = file_path
                    break
            else:
                # NB: the POT file needs the extra `"potcar" not in f` check to ensure that
                # POTCARs are not classed as volumetric files.
                if any(
                    vf.lower() in f and "potcar" not in f
                    for vf in VASP_VOLUMETRIC_FILES
                ):
                    if "volumetric_files" not in by_type[_f.calc_suffix]:
                        by_type[calc_suffix]["volumetric_files"] = []
                    by_type[calc_suffix]["volumetric_files"].append(file_path)  # type: ignore[union-attr]

                elif "poscar.t=" in f:
                    if "elph_poscars" not in by_type[calc_suffix]:
                        by_type[_f.calc_suffix]["elph_poscars"] = []

                    by_type[calc_suffix]["elph_poscars"].append(file_path)  # type: ignore[union-attr]

    return dict(by_type)


def recursive_discover_vasp_files(
    target_dir: FSPathType,
    only_valid: bool = False,
    max_depth: int | None = None,
) -> dict[CalculationLocator, list[FileMetadata]]:
    """
    Recursively scan a target directory and identify VASP files.

    Parameters
    -----------
    target_dir : FSPathType
    only_valid : bool = False (default)
        Whether to only include directories which have the required
        minimum number of input and output files for parsing.
    max_depth : non-negative int or None (default)
        If an int, the maximum depth with which directories are scanned
        for VASP files. For example, if max_depth == 1, this would only
        search `target_dir` and any immediate sub-directories in `target_dir`.

    Returns
    -----------
    dict of Path  to list of FileMetadata identified as VASP files.
    """

    head_dir = Path(target_dir).resolve()

    if max_depth and (max_depth < 0 or not isinstance(max_depth, int)):
        raise ValueError(
            "The maximum path depth should be a non-negative integer, "
            "with zero indicating that only the current directory should "
            "be searched."
        )

    def _path_depth_check(tpath: FSPathType) -> bool:
        if max_depth and (tp := Path(tpath).resolve()) != head_dir:
            for depth, parent in enumerate(tp.parents):
                if parent == head_dir:
                    break
            return depth + 1 <= max_depth
        return True

    def _recursive_discover_vasp_files(
        tdir: FSPathType, paths: dict[CalculationLocator, list[FileMetadata]]
    ) -> None:
        if Path(tdir).is_dir() and _path_depth_check(tdir):
            with os.scandir(tdir) as scan_dir:
                for p in scan_dir:
                    _recursive_discover_vasp_files(p, paths)

            if tpaths_by_suffix := discover_vasp_files(tdir):
                for calc_suffix, tpaths in tpaths_by_suffix.items():
                    # Check if minimum number of VASP files are present
                    # TODO: update with vaspout.h5 parsing
                    if only_valid and any(
                        not any(f in file.name for file in tpaths)
                        for f in REQUIRED_VASP_FILES
                    ):
                        # Incomplete calculation input/output
                        continue
                    paths[
                        CalculationLocator(
                            path=Path(tdir).resolve(), modifier=calc_suffix
                        )
                    ] = tpaths

    vasp_files: dict[CalculationLocator, list[FileMetadata]] = {}
    _recursive_discover_vasp_files(target_dir, vasp_files)
    return vasp_files



from emmet.core.defect import DefectTaskDoc
from emmet.core.testing_utils import DataArchive


def test_parsing_defect_directory(test_dir):
    from pymatgen.analysis.defects.core import Defect

    with DataArchive.extract(test_dir / "vasp/defect_run.json.gz") as temp_dir:
        defect_task = DefectTaskDoc.from_directory(temp_dir)
    assert isinstance(defect_task.defect, Defect)
    assert defect_task.defect_name == "O_Te"



from emmet.core.qchem.calc_types.calc_types import (
    FUNCTIONALS,
    BASIS_SETS,
    SOLVENT_MODELS,
    TASK_TYPES,
)

_REFERENCE_MEMBER_COUNT = {
    "LevelOfTheory": len(FUNCTIONALS) * len(BASIS_SETS) * len(SOLVENT_MODELS),
    "TaskType": len(TASK_TYPES),
}
_REFERENCE_MEMBER_COUNT["CalcType"] = (
    _REFERENCE_MEMBER_COUNT["LevelOfTheory"] * _REFERENCE_MEMBER_COUNT["TaskType"]
)


def test_enums():
    from emmet.core.qchem.calc_types import enums as qchem_enums

    for k, count in _REFERENCE_MEMBER_COUNT.items():
        # Test that we have expected number of enum members
        curr_enum = getattr(qchem_enums, k, None)
        assert len(curr_enum) == count



from typing import Annotated, TypeVar

from pydantic import BeforeValidator, WrapSerializer
from pymatgen.electronic_structure.dos import CompleteDos
from typing_extensions import TypedDict

from emmet.core.types.pymatgen_types.structure_adapter import (
    TypedStructureDict,
    pop_empty_structure_keys,
)

TypedDosDict = TypedDict(
    "TypedDosDict",
    {
        "@module": str,
        "@class": str,
        "efermi": float,
        "energies": list[float],
        "densities": dict[str, list[float]],
    },
)

TypedCompleteDosDict = TypedDict(
    "TypedCompleteDosDict",
    {
        "@module": str,
        "@class": str,
        "efermi": float,
        "structure": TypedStructureDict,
        "energies": list[float],
        "densities": dict[str, list[float]],
        "pdos": list[dict[str, dict[str, dict[str, list[float]]]]],
        "atom_dos": dict[str, TypedDosDict],
        "spd_dos": dict[str, TypedDosDict],
    },
)


CompleteDosTypeVar = TypeVar("CompleteDosTypeVar", CompleteDos, TypedCompleteDosDict)


def pop_empty_dos_keys(dos: CompleteDosTypeVar):
    if isinstance(dos, dict):
        dos["structure"] = pop_empty_structure_keys(dos["structure"])
        return CompleteDos.from_dict(dos)  # type: ignore[arg-type]

    return dos


CompleteDosType = Annotated[
    CompleteDosTypeVar,
    BeforeValidator(pop_empty_dos_keys),
    WrapSerializer(lambda x, nxt, info: x.as_dict(), return_type=TypedCompleteDosDict),
]



from pydantic import BaseModel, Field


class FormulaAutocomplete(BaseModel):
    """
    Class defining formula autocomplete return data
    """

    formula_pretty: str | None = Field(
        None,
        description="Human readable chemical formula.",
    )



from typing import NotRequired

from typing_extensions import TypedDict


class TypedPerIonProps(TypedDict):
    s: float
    p: float
    d: float
    tot: float


TypedOutcarDict = TypedDict(
    "TypedOutcarDict",
    {
        "@module": str,
        "@class": str,
        "efermi": NotRequired[float | None],
        "magnetization": NotRequired[tuple[TypedPerIonProps, ...] | None],
        "charge": NotRequired[tuple[TypedPerIonProps, ...] | None],
        "total_magnetization": NotRequired[float | None],
        "nelect": NotRequired[float | None],
        "is_stopped": NotRequired[bool | None],
        "drift": NotRequired[list[list[float]] | None],
        "ngf": NotRequired[list[int] | None],
        "sampling_radii": NotRequired[list[float] | None],
        "electrostatic_potential": NotRequired[list[float] | None],
    },
)

# would be nice to figure out a constructor for a pmg Outcar from a dict input

# OutcarTypeVar = TypeVar("OutcarTypeVar", Outcar, TypedOutcarDict)

# OutcarType = Annotated[
#     OutcarTypeVar,
#     BeforeValidator(lambda x: Outcar(**x) if isinstance(x, dict) else x),
#     WrapSerializer(lambda x, nxt, info: x.as_dict(), return_type=TypedOutcarDict),
# ]

OutcarType = TypedOutcarDict



from hashlib import blake2b

from pydantic import Field

from emmet.core.molecules import MolPropertyOrigin
from emmet.core.math import Vector3D
from emmet.core.molecules.molecule_property import PropertyDoc
from emmet.core.mpid import MPculeID
from emmet.core.qchem.task import TaskDocument

__author__ = "Evan Spotte-Smith <ewcspottesmith@lbl.gov>"


class ElectricMultipoleDoc(PropertyDoc):
    """Electric multipole (dipole, etc.) properties of a molecule"""

    property_name: str = "multipole_moments"

    total_dipole: float = Field(
        ...,
        description="Total molecular dipole moment (Debye)",
    )

    dipole_moment: Vector3D = Field(
        ...,
        description="Molecular dipole moment vector (Debye)",
    )

    resp_total_dipole: float | None = Field(
        None,
        description="Total dipole moment, calculated via restrained electrostatic potential (RESP) (Debye)",
    )

    resp_dipole_moment: Vector3D | None = Field(
        None,
        description="Molecular dipole moment vector, calculated via RESP (Debye)",
    )

    quadrupole_moment: dict[str, float] | None = Field(
        None,
        description="Quadrupole moment components (Debye Ang)",
    )

    octopole_moment: dict[str, float] | None = Field(
        None,
        description="Octopole moment components (Debye Ang^2)",
    )

    hexadecapole_moment: dict[str, float] | None = Field(
        None,
        description="Hexadecapole moment tensor components (Debye Ang^2)",
    )

    @classmethod
    def from_task(
        cls,
        task: TaskDocument,
        molecule_id: MPculeID,
        deprecated: bool = False,
        **kwargs,
    ):  # type: ignore[override]
        """
        Construct an electric multipole doc
        """

        if task.output.optimized_molecule is not None:
            mol = task.output.optimized_molecule
        else:
            mol = task.output.initial_molecule

        task_type = task.task_type.value

        # For SP and force calcs, dipoles stored in output
        dipoles = task.output.dipoles
        if dipoles is None:
            if task_type in ["Single Point", "Force"]:
                raise Exception("No dipoles in task!")
            total_dipole = None
            dipole_moment = None
            resp_total_dipole = None
            resp_dipole_moment = None
        else:
            total_dipole = dipoles.get("total")
            dipole_moment = dipoles.get("dipole")
            resp_total_dipole = dipoles.get("RESP_total")
            resp_dipole_moment = dipoles.get("RESP_dipole")

        # Look for multipole moment in calcs_reversed
        calcs_reversed = task.calcs_reversed

        # FFOpt/FFTSOpt: grab final multipole moments from final optimization
        if task_type in [
            "Frequency Flattening Geometry Optimization",
            "Frequency Flattening Transition State Geometry Optimization",
        ]:
            calc = calcs_reversed[1]
            grab_index = -1
        # Frequency calc: grab first multipole moments
        # This is necessary if finite difference is used, in which case multipole moments will be generated for each
        # perturbation
        elif task_type == "Frequency Analysis":
            calc = calcs_reversed[0]
            grab_index = 0
        # Otherwise, grab final available multipole moments
        else:
            calc = calcs_reversed[0]
            grab_index = -1

        dipoles = calc.get("dipoles", dict())

        if total_dipole is None:
            total_dipole = dipoles.get("total")
            if isinstance(total_dipole, list):
                total_dipole = total_dipole[grab_index]
        if dipole_moment is None:
            dipole_moment = dipoles.get("dipole")
            if isinstance(dipole_moment, list) and len(dipole_moment) == 0:
                dipole_moment = None
            elif isinstance(dipole_moment[0], list):
                dipole_moment = dipole_moment[grab_index]
        if resp_total_dipole is None:
            resp_total_dipole = dipoles.get("RESP_total")
            if isinstance(resp_total_dipole, list):
                resp_total_dipole = resp_total_dipole[grab_index]
        if resp_dipole_moment is None:
            resp_dipole_moment = dipoles.get("RESP_dipole")
            if isinstance(resp_dipole_moment, list) and len(resp_dipole_moment) == 0:
                resp_dipole_moment = None
            elif isinstance(resp_dipole_moment[0], list):
                resp_dipole_moment = resp_dipole_moment[grab_index]

        if total_dipole is None or dipole_moment is None:
            raise Exception("Total dipole or dipole moment vector missing!")

        multipoles = calc.get("multipoles", dict())
        quadrupoles = multipoles.get("quadrupole")
        octopoles = multipoles.get("octopole")
        hexadecapoles = multipoles.get("hexadecapole")

        quadrupole_moment = None
        octopole_moment = None
        hexadecapole_moment = None

        if quadrupoles is not None:
            if isinstance(quadrupoles, dict):
                quadrupole_moment = quadrupoles
            else:
                quadrupole_moment = quadrupoles[grab_index]

        if octopoles is not None:
            if isinstance(octopoles, dict):
                octopole_moment = octopoles
            else:
                octopole_moment = octopoles[grab_index]

        if hexadecapoles is not None:
            if isinstance(hexadecapoles, dict):
                hexadecapole_moment = hexadecapoles
            else:
                hexadecapole_moment = hexadecapoles[grab_index]

        id_string = f"electricmultipole-{molecule_id}-{task.task_id}-{task.lot_solvent}"
        h = blake2b()
        h.update(id_string.encode("utf-8"))
        property_id = h.hexdigest()

        return super().from_molecule(
            meta_molecule=mol,
            property_id=property_id,
            molecule_id=molecule_id,
            level_of_theory=task.level_of_theory,
            solvent=task.solvent,
            lot_solvent=task.lot_solvent,
            molecule=mol,
            total_dipole=total_dipole,
            dipole_moment=dipole_moment,
            resp_total_dipole=resp_total_dipole,
            resp_dipole_moment=resp_dipole_moment,
            quadrupole_moment=quadrupole_moment,
            octopole_moment=octopole_moment,
            hexadecapole_moment=hexadecapole_moment,
            warnings=list(),  # No warnings currently available
            origins=[MolPropertyOrigin(name="multipole_moments", task_id=task.task_id)],
            deprecated=deprecated,
            **kwargs,
        )



"""Core definition of a CorrectedEntriesDoc Document"""

from pydantic import Field

from emmet.core.base import EmmetBaseModel
from emmet.core.types.enums import ThermoType
from emmet.core.types.pymatgen_types.computed_entries_adapter import (
    ComputedStructureEntryType,
)
from emmet.core.utils import type_override
from emmet.core.types.typing import DateTimeType
from emmet.core.vasp.calc_types.enums import RunType


@type_override({"entries": dict[ThermoType, list[ComputedStructureEntryType]]})
class CorrectedEntriesDoc(EmmetBaseModel):
    """
    A corrected entries document
    """

    property_name: str = "corrected_entries"

    chemsys: str = Field(
        ...,
        title="Chemical System",
        description="Dash-delimited string of elements in the material.",
    )

    entries: dict[RunType | ThermoType, list[ComputedStructureEntryType] | None] = (
        Field(
            ...,
            description="List of all corrected entries that are valid for the specified thermo type.",
        )
    )

    last_updated: DateTimeType = Field(
        description="Timestamp for the most recent calculation update for this property.",
    )



from typing import Annotated, TypeVar

from pydantic import BeforeValidator, WrapSerializer
from pymatgen.core import Structure
from pymatgen.core.structure import Molecule
from typing_extensions import NotRequired, TypedDict

from emmet.core.types.pymatgen_types.lattice_adapter import TypedLatticeDict
from emmet.core.types.pymatgen_types.properties import TypedAggregateProperitesDict
from emmet.core.types.pymatgen_types.sites_adapter import TypedSiteDict

TypedStructureDict = TypedDict(
    "TypedStructureDict",
    {
        "@module": str,
        "@class": str,
        "charge": NotRequired[float | None],
        "lattice": TypedLatticeDict,
        "sites": list[TypedSiteDict],
        "properties": NotRequired[TypedAggregateProperitesDict | None],
    },
)

StructureTypeVar = TypeVar("StructureTypeVar", Structure, TypedStructureDict)

TypedMoleculeDict = TypedDict(
    "TypedMoleculeDict",
    {
        "@module": str,
        "@class": str,
        "charge": float,
        "spin_multiplicity": int,
        "sites": list[TypedSiteDict],
        "properties": TypedAggregateProperitesDict,
    },
)


MoleculeTypeVar = TypeVar("MoleculeTypeVar", Molecule, TypedMoleculeDict)


def pop_empty_structure_keys(inp: StructureTypeVar | MoleculeTypeVar):
    if isinstance(inp, dict):
        target_cls = Structure if inp["@class"] == "Structure" else Molecule

        if inp.get("properties"):
            for prop, val in list(inp["properties"].items()):  # type: ignore[union-attr]
                if val is None:
                    del inp["properties"][prop]  # type: ignore[union-attr]

        for site in inp["sites"]:
            if "name" in site:
                if not site["name"]:
                    del site["name"]

            if site.get("properties"):
                for prop, val in list(site["properties"].items()):
                    if val is None:
                        del site["properties"][prop]

            for species in site["species"]:
                for prop, val in list(species.items()):
                    if val is None:
                        del species[prop]

        return target_cls.from_dict(inp)  # type: ignore[arg-type]

    return inp


MoleculeType = Annotated[
    MoleculeTypeVar,
    BeforeValidator(pop_empty_structure_keys),
    WrapSerializer(lambda x, nxt, info: x.as_dict(), return_type=TypedMoleculeDict),
]

StructureType = Annotated[
    StructureTypeVar,
    BeforeValidator(pop_empty_structure_keys),
    WrapSerializer(lambda x, nxt, info: x.as_dict(), return_type=TypedStructureDict),
]



import pytest
import numpy as np
from monty.io import zopen
import json
from monty.json import MontyDecoder
import importlib

from emmet.core.testing_utils import TEST_FILES_DIR

OPENBABEL_INSTALLED = importlib.util.find_spec("openbabel") is not None


@pytest.fixture(scope="session")
def test_dir():
    return TEST_FILES_DIR


class SchemaTestData:
    """Dummy class to be used to contain all test data information"""


class SinglePointTest(SchemaTestData):
    folder = "qchem_sp_test"
    task_files = {
        "standard": {
            "qcinput_file": "mol.qin.gz",
            "qcoutput_file": "mol.qout.gz",
        }
    }

    objects = {"standard": []}
    task_doc = {
        "calcs_reversed": [
            {
                "output": {
                    "mulliken": [np.array([-0.713178, 0.357278, 0.3559])],
                    "resp": [np.array([-0.872759, 0.436379, 0.436379])],
                    "final_energy": -76.4493700739,
                },
                "input": {
                    "charge": 0,
                    "rem": {
                        "job_type": "sp",
                        "basis": "def2-qzvppd",
                        "max_scf_cycles": "100",
                        "gen_scfman": "true",
                        "xc_grid": "3",
                        "thresh": "14",
                        "s2thresh": "16",
                        "scf_algorithm": "diis",
                        "resp_charges": "true",
                        "symmetry": "false",
                        "sym_ignore": "true",
                        "method": "wb97mv",
                        "solvent_method": "smd",
                        "ideriv": "1",
                    },
                    "job_type": "sp",
                },
            }
        ],
        "input": {
            "initial_molecule": {
                "@module": "pymatgen.core.structure",
                "@class": "Molecule",
                "charge": 0.0,
                "spin_multiplicity": 1,
                "sites": [
                    {
                        "name": "O",
                        "species": [{"element": "O", "occu": 1}],
                        "xyz": [-0.80595, 2.22952, -0.01914],
                        "properties": {},
                        "label": "O",
                    },
                    {
                        "name": "H",
                        "species": [{"element": "H", "occu": 1}],
                        "xyz": [0.18338, 2.20176, 0.01351],
                        "properties": {},
                        "label": "H",
                    },
                    {
                        "name": "H",
                        "species": [{"element": "H", "occu": 1}],
                        "xyz": [-1.09531, 1.61602, 0.70231],
                        "properties": {},
                        "label": "H",
                    },
                ],
            },
            "rem": {
                "job_type": "sp",
                "basis": "def2-qzvppd",
                "max_scf_cycles": "100",
                "gen_scfman": "true",
                "xc_grid": "3",
                "thresh": "14",
                "s2thresh": "16",
                "scf_algorithm": "diis",
                "resp_charges": "true",
                "symmetry": "false",
                "sym_ignore": "true",
                "method": "wb97mv",
                "solvent_method": "smd",
                "ideriv": "1",
            },
            "level_of_theory": "wB97M-V/def2-QZVPPD/SMD",
            "task_type": "Single Point",
            "calc_type": "wB97M-V/def2-QZVPPD/SMD Single Point",
            "solvation_lot_info": "wB97M-V/def2-QZVPPD/SMD(SOLVENT=WATER)",
        },
        "output": {
            "mulliken": [np.array([-0.713178, 0.357278, 0.3559])],
            "resp": [np.array([-0.872759, 0.436379, 0.436379])],
            "final_energy": -76.4493700739,
            "dipoles": {
                "total": 2.4648,
                "dipole": [np.array([1.4227, -1.3039, 1.5333])],
                "RESP_total": 2.5411,
                "RESP_dipole": [np.array([1.4672, -1.3441, 1.5806])],
            },
        },
        "custodian": [
            {
                "job": {
                    "@module": "custodian.qchem.jobs",
                    "@class": "QCJob",
                    "@version": "2022.5.26",
                    "qchem_command": ["qchem"],
                    "max_cores": "40",
                    "multimode": "openmp",
                    "input_file": "mol.qin",
                    "output_file": "mol.qout",
                    "qclog_file": "mol.qclog",
                    "suffix": "",
                    "calc_loc": "/tmp",
                    "nboexe": None,
                    "save_scratch": False,
                    "backup": "true",
                },
                "corrections": [],
            }
        ],
    }


class OptimizationTest(SchemaTestData):
    folder = "qchem_opt_test"
    task_files = {
        "standard": {
            "qcinput_file": "mol.qin.gz",
            "qcoutput_file": "mol.qout.gz",
        }
    }

    objects = {"standard": []}
    task_doc = {
        "calcs_reversed": [
            {
                "output": {
                    "optimized_molecule": {
                        "@module": "pymatgen.core.structure",
                        "@class": "Molecule",
                        "charge": 0,
                        "spin_multiplicity": 1,
                        "sites": [
                            {
                                "name": "O",
                                "species": [{"element": "O", "occu": 1}],
                                "xyz": [-0.8008592596, 2.2248298937, -0.0136245943],
                                "properties": {},
                                "label": "O",
                            },
                            {
                                "name": "H",
                                "species": [{"element": "H", "occu": 1}],
                                "xyz": [0.1637955748, 2.1962925542, 0.0199393927],
                                "properties": {},
                                "label": "H",
                            },
                            {
                                "name": "H",
                                "species": [{"element": "H", "occu": 1}],
                                "xyz": [-1.0808163152, 1.6261775521, 0.6903652017],
                                "properties": {},
                                "label": "H",
                            },
                        ],
                    },
                    "mulliken": [-0.373491, 0.186964, 0.186527],
                    "resp": [-0.89522, 0.44761, 0.44761],
                    "final_energy": -76.358341626913,
                },
                "input": {
                    "charge": 0,
                    "rem": {
                        "job_type": "sp",
                        "basis": "def2-qzvppd",
                        "max_scf_cycles": "100",
                        "gen_scfman": "true",
                        "xc_grid": "3",
                        "thresh": "14",
                        "s2thresh": "16",
                        "scf_algorithm": "diis",
                        "resp_charges": "true",
                        "symmetry": "false",
                        "sym_ignore": "true",
                        "method": "wb97mv",
                        "solvent_method": "smd",
                        "ideriv": "1",
                    },
                    "job_type": "sp",
                },
            }
        ],
        "input": {
            "initial_molecule": {
                "@module": "pymatgen.core.structure",
                "@class": "Molecule",
                "charge": 0.0,
                "spin_multiplicity": 1,
                "sites": [
                    {
                        "name": "O",
                        "species": [{"element": "O", "occu": 1}],
                        "xyz": [-0.80595, 2.22952, -0.01914],
                        "properties": {},
                        "label": "O",
                    },
                    {
                        "name": "H",
                        "species": [{"element": "H", "occu": 1}],
                        "xyz": [0.18338, 2.20176, 0.01351],
                        "properties": {},
                        "label": "H",
                    },
                    {
                        "name": "H",
                        "species": [{"element": "H", "occu": 1}],
                        "xyz": [-1.09531, 1.61602, 0.70231],
                        "properties": {},
                        "label": "H",
                    },
                ],
            },
            "rem": {
                "job_type": "opt",
                "basis": "def2-svpd",
                "max_scf_cycles": "100",
                "gen_scfman": "true",
                "xc_grid": "3",
                "thresh": "14",
                "s2thresh": "16",
                "scf_algorithm": "diis",
                "resp_charges": "true",
                "symmetry": "false",
                "sym_ignore": "true",
                "method": "wb97mv",
                "solvent_method": "smd",
                "ideriv": "1",
                "geom_opt2": "3",
            },
            "level_of_theory": "wB97M-V/def2-SVPD/SMD",
            "task_type": "Geometry Optimization",
            "calc_type": "wB97M-V/def2-SVPD/SMD Geometry Optimization",
            "solvation_lot_info": "wB97M-V/def2-SVPD/SMD(SOLVENT=WATER)",
        },
        "output": {
            "initial_molecule": {
                "@module": "pymatgen.core.structure",
                "@class": "Molecule",
                "charge": 0.0,
                "spin_multiplicity": 1,
                "sites": [
                    {
                        "name": "O",
                        "species": [{"element": "O", "occu": 1}],
                        "xyz": [-0.80595, 2.22952, -0.01914],
                        "properties": {},
                        "label": "O",
                    },
                    {
                        "name": "H",
                        "species": [{"element": "H", "occu": 1}],
                        "xyz": [0.18338, 2.20176, 0.01351],
                        "properties": {},
                        "label": "H",
                    },
                    {
                        "name": "H",
                        "species": [{"element": "H", "occu": 1}],
                        "xyz": [-1.09531, 1.61602, 0.70231],
                        "properties": {},
                        "label": "H",
                    },
                ],
            },
            "optimized_molecule": {
                "@module": "pymatgen.core.structure",
                "@class": "Molecule",
                "charge": 0,
                "spin_multiplicity": 1,
                "sites": [
                    {
                        "name": "O",
                        "species": [{"element": "O", "occu": 1}],
                        "xyz": [-0.8008592596, 2.2248298937, -0.0136245943],
                        "properties": {},
                        "label": "O",
                    },
                    {
                        "name": "H",
                        "species": [{"element": "H", "occu": 1}],
                        "xyz": [0.1637955748, 2.1962925542, 0.0199393927],
                        "properties": {},
                        "label": "H",
                    },
                    {
                        "name": "H",
                        "species": [{"element": "H", "occu": 1}],
                        "xyz": [-1.0808163152, 1.6261775521, 0.6903652017],
                        "properties": {},
                        "label": "H",
                    },
                ],
            },
            "mulliken": [-0.373491, 0.186964, 0.186527],
            "resp": [-0.89522, 0.44761, 0.44761],
            "final_energy": -76.358341626913,
        },
        "custodian": [
            {
                "job": {
                    "@module": "custodian.qchem.jobs",
                    "@class": "QCJob",
                    "@version": "2022.5.26",
                    "qchem_command": ["qchem"],
                    "max_cores": "40",
                    "multimode": "openmp",
                    "input_file": "mol.qin",
                    "output_file": "mol.qout",
                    "qclog_file": "mol.qclog",
                    "suffix": "",
                    "calc_loc": "/tmp",
                    "nboexe": "null",
                    "save_scratch": "false",
                    "backup": "true",
                },
                "corrections": [],
            }
        ],
    }


objects = {cls.__name__: cls for cls in SchemaTestData.__subclasses__()}


def get_test_object(object_name):
    """Get the schema test data object from the class name."""
    return objects[object_name]


def safe_load(path):
    with zopen(path, "rt") as f:
        data = json.load(f)
    if not OPENBABEL_INSTALLED:
        data.pop("custodian", None)
    return MontyDecoder().process_decoded(data)



import numpy as np
import pytest
from monty.serialization import loadfn
from pymatgen.core import Structure

from emmet.core import ARROW_COMPATIBLE
from emmet.core.absorption import AbsorptionDoc
from emmet.core.utils import jsanitize

if ARROW_COMPATIBLE:
    import pyarrow as pa

    from emmet.core.arrow import arrowize


@pytest.fixture(scope="session")
def absorption_test_data(test_dir):
    return loadfn(test_dir / "sample_absorptions.json.gz")


@pytest.fixture(scope="module")
def absorption_test_doc(absorption_test_data):
    structure = Structure.from_dict(
        jsanitize(absorption_test_data["input"]["structure"])
    )
    task_id = absorption_test_data["task_id"]
    kpoints = absorption_test_data["orig_inputs"]["kpoints"]

    cr = absorption_test_data["calcs_reversed"][0]["output"]
    doc = AbsorptionDoc.from_structure(
        structure=structure,
        material_id="mp-{}".format(task_id),
        task_id=task_id,
        deprecated=False,
        energies=cr["frequency_dependent_dielectric"]["energy"],
        real_d=cr["frequency_dependent_dielectric"]["real"],
        imag_d=cr["frequency_dependent_dielectric"]["imaginary"],
        absorption_co=cr["optical_absorption_coeff"],
        bandgap=absorption_test_data["output"]["bandgap"],
        nkpoints=kpoints.num_kpts,
        is_hubbard=False,
    )

    return doc


def test_absorption_doc(absorption_test_doc):
    absorption_coeff = np.array(
        [
            0,
            5504161142.509,
            16485480924.4252,
            41235259342.4927,
            76990619286.8861,
            109929386572.273,
            164921201202.527,
            230913400825.579,
            285790460873.292,
            371002598552.062,
        ]
    )

    imag_dielectric = [0.0, 0.0002, 0.0003]

    energies = [
        0,
        0.0309,
        0.0617,
        0.0926,
        0.1235,
        0.1543,
        0.1852,
        0.2161,
        0.2469,
        0.2778,
    ]

    assert absorption_test_doc is not None
    assert absorption_test_doc.property_name == "Optical absorption spectrum"
    assert absorption_test_doc.energies[0:10] == energies
    assert absorption_test_doc.material_id == "mp-1316"
    assert absorption_test_doc.absorption_coefficient[0:10] == list(absorption_coeff)
    assert absorption_test_doc.average_imaginary_dielectric[0:3] == imag_dielectric
    assert absorption_test_doc.bandgap == 4.4652


@pytest.mark.skipif(
    not ARROW_COMPATIBLE, reason="pyarrow must be installed to run this test."
)
def test_arrow(absorption_test_doc):
    arrow_struct = pa.scalar(
        absorption_test_doc.model_dump(context={"format": "arrow"}),
        type=arrowize(AbsorptionDoc),
    )

    test_arrow_doc = AbsorptionDoc(**arrow_struct.as_py(maps_as_pydicts="strict"))

    assert absorption_test_doc.model_dump() == test_arrow_doc.model_dump()



import importlib
import inspect
import itertools
import os
from pathlib import Path
from typing import Optional, Union

import pytest
from pydantic._internal._model_construction import ModelMetaclass

from emmet.core import ARROW_COMPATIBLE

pa = pytest.importorskip("pyarrow")

if ARROW_COMPATIBLE:
    from emmet.core.arrow import arrowize


def import_models():
    core_path = Path(__file__).parent.parent.joinpath("emmet/core")
    core_models = []
    for root, dirs, files in os.walk(core_path):
        if "__pycache__" in dirs:
            dirs.remove("__pycache__")

        parent_module = ".".join(
            [
                "emmet",
                "core",
                *list(
                    itertools.takewhile(
                        lambda x: x != "core", reversed(root.split("/"))
                    )
                )[::-1],
            ]
        )

        for file in files:
            if file not in ["__init__.py", "__pycache__"]:
                if file[-3:] != ".py":
                    continue

                file_name = file[:-3]
                if file_name == "optimade":
                    continue
                module_name = f"{parent_module}.{file_name}"

                for name, obj in inspect.getmembers(
                    importlib.import_module(module_name), inspect.isclass
                ):
                    if (
                        obj.__module__ == module_name
                        and isinstance(obj, ModelMetaclass)
                        and not hasattr(obj, "arrow_incompatible")
                    ):
                        core_models.append(obj)

    return core_models


def test_arrowize_succeeds():
    assert all(
        dtype == pa.int64()
        for dtype in [
            arrowize(int | None),
            arrowize(Optional[int]),
            arrowize(Union[int | None]),
        ]
    )

    assert all(
        dtype == pa.map_(pa.string(), pa.int64())
        for dtype in [arrowize(dict[str, int]), arrowize(dict[str, int | None])]
    )


def test_arrowize_fails():
    with pytest.raises(AssertionError):
        arrowize(dict)
        arrowize(list)
        arrowize(tuple)


@pytest.mark.parametrize("model", import_models())
def test_document_models_for_arrow_compatibility(model):
    assert isinstance(arrowize(model), pa.DataType)



"""Settings for defaults in the core definitions of Materials Project Documents."""

from __future__ import annotations

import gzip
import json
from pathlib import Path
from typing import TYPE_CHECKING

import requests  # type: ignore[import-untyped]
from monty.io import zopen
from monty.json import MontyDecoder
from pydantic import Field, ImportString, field_validator, model_validator
from pydantic_settings import BaseSettings, SettingsConfigDict

if TYPE_CHECKING:
    from typing import Any
    from typing_extensions import Self

DEFAULT_CONFIG_FILE_PATH = str(Path.home().joinpath(".emmet.json"))


class EmmetSettings(BaseSettings):
    """
    Settings for the emmet- packages
    Non-core packages should subclass this to get settings specific to their needs
    The default way to modify these is to modify ~/.emmet.json or set the environment variable
    EMMET_CONFIG_FILE to point to the json with emmet settings
    """

    config_file: str = Field(
        DEFAULT_CONFIG_FILE_PATH, description="File to load alternative defaults from"
    )

    LTOL: float = Field(
        0.2, description="Fractional length tolerance for structure matching"
    )
    STOL: float = Field(
        0.3,
        description="Site tolerance for structure matching. Defined as the fraction of the"
        " average free length per atom = ( V / Nsites ) ** (1/3)",
    )
    SYMPREC: float = Field(
        0.1, description="Symmetry precision for spglib symmetry finding"
    )
    ANGLE_TOL: float = Field(
        5, description="Angle tolerance for structure matching in degrees."
    )

    PGATOL: float = Field(
        0.3,
        description="Distance tolerance to consider sites as symmetrically equivalent.",
    )
    PGAEIGENTOL: float = Field(
        0.01, description="Tolerance to compare eigen values of the inertia tensor."
    )
    PGAMATRIXTOL: float = Field(
        0.1,
        description="Tolerance used to generate the full set of symmetry operations of the point group.",
    )

    MAX_PIEZO_MILLER: int = Field(
        10,
        description="Maximum miller allowed for computing strain direction for maximal piezo response",
    )

    QCHEM_FUNCTIONAL_QUALITY_SCORES: dict[str, int] = Field(
        {
            "wB97M-V": 7,
            "wB97X-V": 6,
            "wB97X-D3": 5,
            "wB97X-D": 5,
            "B3LYP": 4,
            "B97M-V": 3,
            "B97M-rV": 3,
            "B97-D3": 2,
            "B97-D": 2,
            "PBE": 1,
        },
        description="Dictionary mapping Q-Chem density functionals to a quality score.",
    )

    QCHEM_BASIS_QUALITY_SCORES: dict[str, int] = Field(
        {
            "6-31g*": 1,
            "def2-SVPD": 2,
            "def2-TZVP": 3,
            "def2-TZVPD": 4,
            "def2-TZVPP": 5,
            "def2-TZVPPD": 6,
            "def2-QZVPPD": 7,
        },
        description="Dictionary mapping Q-Chem basis sets to a quality score.",
    )

    QCHEM_SOLVENT_MODEL_QUALITY_SCORES: dict[str, int] = Field(
        {"CMIRS": 7, "SMD": 5, "ISOSVP": 4, "PCM": 3, "VACUUM": 1},
        description="Dictionary mapping Q-Chem solvent models to a quality score.",
    )

    QCHEM_TASK_QUALITY_SCORES: dict[str, int] = Field(
        {
            "single_point": 1,
            "geometry optimization": 2,
            "frequency-flattening geometry optimization": 3,
        },
        description="Dictionary mapping Q-Chem task type to a quality score",
    )

    VASP_STRUCTURE_QUALITY_SCORES: dict[str, int] = Field(
        {"r2SCAN": 5, "SCAN": 4, "GGA+U": 3, "GGA": 2, "PBEsol": 1},
        description="Dictionary Mapping VASP calculation run types to rung level for VASP materials builder structure data",  # noqa: E501
    )

    VASP_KPTS_TOLERANCE: float = Field(
        0.9,
        description="Relative tolerance for kpt density to still be a valid task document",
    )

    VASP_KSPACING_TOLERANCE: float = Field(
        0.05,
        description="Relative tolerance for kspacing to still be a valid task document",
    )

    VASP_MAX_MAGMOM: dict[str, float] = Field(
        {"Cr": 5}, description="Maximum permitted magnetic moments by element type."
    )

    VASP_DEFAULT_INPUT_SETS: dict[str, ImportString] = Field(
        {
            "GGA Structure Optimization": "pymatgen.io.vasp.sets.MPRelaxSet",
            "GGA+U Structure Optimization": "pymatgen.io.vasp.sets.MPRelaxSet",
            "r2SCAN Structure Optimization": "pymatgen.io.vasp.sets.MPScanRelaxSet",
            "SCAN Structure Optimization": "pymatgen.io.vasp.sets.MPScanRelaxSet",
            "PBEsol Structure Optimization": "pymatgen.io.vasp.sets.MPScanRelaxSet",
            "GGA Static": "pymatgen.io.vasp.sets.MPStaticSet",
            "GGA+U Static": "pymatgen.io.vasp.sets.MPStaticSet",
            "r2SCAN Static": "pymatgen.io.vasp.sets.MPScanStaticSet",
            "SCAN Static": "pymatgen.io.vasp.sets.MPScanStaticSet",
            "PBEsol Static": "pymatgen.io.vasp.sets.MPScanStaticSet",
            "HSE06 Static": "pymatgen.io.vasp.sets.MPScanStaticSet",
            "GGA NSCF Uniform": "pymatgen.io.vasp.sets.MPNonSCFSet",
            "GGA+U NSCF Uniform": "pymatgen.io.vasp.sets.MPNonSCFSet",
            "GGA NSCF Line": "pymatgen.io.vasp.sets.MPNonSCFSet",
            "GGA+U NSCF Line": "pymatgen.io.vasp.sets.MPNonSCFSet",
            "GGA NMR Electric Field Gradient": "pymatgen.io.vasp.sets.MPNMRSet",
            "GGA NMR Nuclear Shielding": "pymatgen.io.vasp.sets.MPNMRSet",
            "GGA+U NMR Electric Field Gradient": "pymatgen.io.vasp.sets.MPNMRSet",
            "GGA+U NMR Nuclear Shielding": "pymatgen.io.vasp.sets.MPNMRSet",
            "GGA Deformation": "pymatgen.io.vasp.sets.MPStaticSet",
            "GGA+U Deformation": "pymatgen.io.vasp.sets.MPStaticSet",
            "GGA DFPT Dielectric": "pymatgen.io.vasp.sets.MPStaticSet",
            "GGA+U DFPT Dielectric": "pymatgen.io.vasp.sets.MPStaticSet",
        },
        description="Default input sets for task validation",
    )

    VASP_VALIDATE_POTCAR_STATS: bool = Field(
        True, description="Whether to validate POTCAR stat values."
    )

    VASP_CHECKED_LDAU_FIELDS: list[str] = Field(
        ["LDAUU", "LDAUJ", "LDAUL"], description="LDAU fields to validate for tasks"
    )

    VASP_MAX_SCF_GRADIENT: float = Field(
        100,
        description="Maximum upward gradient in the last SCF for any VASP calculation",
    )

    VASP_USE_STATICS: bool = Field(
        True,
        description="Use static calculations for structure and energy along with structure optimizations",
    )
    model_config = SettingsConfigDict(env_prefix="emmet_", extra="ignore")

    @model_validator(mode="before")
    @classmethod
    def load_default_settings(cls, values: Any) -> Any:
        """
        Loads settings from a root file if available and uses that as defaults in
        place of built in defaults
        """
        config_file_path: str = values.get("config_file", DEFAULT_CONFIG_FILE_PATH)

        new_values = {}

        # TODO: do we want to support gzipped config files?
        if config_file_path.startswith("http"):
            response = requests.get(config_file_path)
            if response.content.startswith(b"\x1f\x8b"):
                new_values = json.loads(gzip.decompress(response.content))
            else:
                new_values = json.loads(response.content.decode())
        elif Path(config_file_path).exists():
            with zopen(config_file_path, "rt") as f:
                new_values = json.load(f)

        new_values.update(values)

        return new_values

    @classmethod
    def autoload(cls, settings: None | dict | Self) -> Self:
        if settings is None:
            return cls(**{})
        elif isinstance(settings, dict):
            return cls(**settings)
        return settings

    @field_validator("VASP_DEFAULT_INPUT_SETS", mode="before")
    @classmethod
    def convert_input_sets(cls, value):
        if isinstance(value, dict):
            return {k: MontyDecoder().process_decoded(v) for k, v in value.items()}
        return value

    def as_dict(self):
        """
        HotPatch to enable serializing EmmetSettings via Monty
        """
        return self.model_dump(exclude_unset=True, exclude_defaults=True)



from typing_extensions import NotRequired, TypedDict

from emmet.core.math import Vector3D


class TypedSiteProperitesDict(TypedDict):
    magmom: NotRequired[float | None]
    charge: NotRequired[float | None]
    velocities: NotRequired[Vector3D | None]
    selective_dynamics: NotRequired[tuple[bool, bool, bool] | None]
    coordination_no: NotRequired[int | None]
    forces: NotRequired[list[float, float, float] | None]  # type: ignore[type-arg]


class TypedAggregateProperitesDict(TypedDict):
    magmom: NotRequired[list[float] | None]
    charge: NotRequired[list[float] | None]
    velocities: NotRequired[list[Vector3D] | None]
    selective_dynamics: NotRequired[list[tuple[bool, bool, bool]] | None]
